"""
Rule number: '8565'
Module: CIN Details
Rule details: If <CINclosureDate> (N00102) is present then it must be on or after all of the following dates that are present:
    <AssessmentActualStartDate> (N00159)
    <AssessmentAuthorisationDate>(N00160)
    <S47ActualStartDate> (N00148)
    <DateOfInitialCPC> (N00110)
    <CPPendDate> (N00115)
    <CINPlanStartDate> (N00689)
    <CINPlanEndDate> (N00690)
Rule message: Activity shown after a case has been closed

"""
from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

CINdetails = CINTable.CINdetails
LAchildID = CINdetails.LAchildID
CINdetailsID = CINdetails.CINdetailsID
CINclosureDate = CINdetails.CINclosureDate
DateOfInitialCPC = CINdetails.DateOfInitialCPC

Assessments = CINTable.Assessments
LAchildID = Assessments.LAchildID
CINdetailsID = Assessments.CINdetailsID
AssessmentActualStartDate = Assessments.AssessmentActualStartDate
AssessmentAuthorisationDate = Assessments.AssessmentAuthorisationDate

Section47 = CINTable.Section47
LAchildID = Section47.LAchildID
CINdetailsID = Section47.CINdetailsID
S47ActualStartDate = Section47.S47ActualStartDate

ChildProtectionPlans = CINTable.ChildProtectionPlans
LAchildID = ChildProtectionPlans.LAchildID
CINdetailsID = ChildProtectionPlans.CINdetailsID
CPPendDate = ChildProtectionPlans.CPPendDate

CINplanDates = CINTable.CINplanDates
LAchildID = CINplanDates.LAchildID
CINdetailsID = CINplanDates.CINdetailsID
CINPlanStartDate = CINplanDates.CINPlanStartDate
CINPlanEndDate = CINplanDates.CINPlanEndDate


@rule_definition(
    code="8565",
    module=CINTable.ChildProtectionPlans,
    message="Activity shown after a case has been closed",
    affected_fields=[
        CINclosureDate,
        DateOfInitialCPC,
        AssessmentActualStartDate,
        AssessmentAuthorisationDate,
        S47ActualStartDate,
        CPPendDate,
        CINPlanStartDate,
        CINPlanEndDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_cin = data_container[CINdetails]
    df_ass = data_container[Assessments]
    df_47 = data_container[Section47]
    df_cpp = data_container[ChildProtectionPlans]
    df_plan = data_container[CINplanDates]

    df_cin.index.name = "ROW_ID"
    df_ass.index.name = "ROW_ID"
    df_47.index.name = "ROW_ID"
    df_cpp.index.name = "ROW_ID"
    df_plan.index.name = "ROW_ID"

    df_cin.reset_index(inplace=True)
    df_ass.reset_index(inplace=True)
    df_47.reset_index(inplace=True)
    df_cpp.reset_index(inplace=True)
    df_plan.reset_index(inplace=True)

    # Rule details: If <CINclosureDate> (N00102) is present then it must be on or after all of the following dates that are present:
    #     <AssessmentActualStartDate> (N00159)
    #     <AssessmentAuthorisationDate>(N00160)
    #     <S47ActualStartDate> (N00148)
    #     <DateOfInitialCPC> (N00110)
    #     <CPPendDate> (N00115)
    #     <CINPlanStartDate> (N00689)
    #     <CINPlanEndDate> (N00690)

    # Remove rows without a CIN closure date
    df_cin = df_cin[df_cin[CINclosureDate].notna()]

    # CIN TABLE
    df_cin["CINclosureDate"] = pd.to_datetime(
        df_cin["CINclosureDate"], format="%d/%m/%Y", errors="coerce"
    )
    df_cin["DateOfInitialCPC"] = pd.to_datetime(
        df_cin["DateOfInitialCPC"], format="%d/%m/%Y", errors="coerce"
    )

    df_cin_fail = df_cin[df_cin["CINclosureDate"] < df_cin["DateOfInitialCPC"]]
    df_cin_fail["ERROR_ID"] = tuple(
        zip(
            df_cin_fail["LAchildID"],
            df_cin_fail["CINdetailsID"],
            df_cin_fail["CINclosureDate"],
        )
    )

    df_cin_issues_cin = (
        df_cin_fail.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # ASSESSMENTS TABLE
    df_cin_ass = df_cin.merge(
        df_ass,
        on=["LAchildID", "CINdetailsID"],
        how="left",
        suffixes=["_cin", "_ass"],
    )

    df_cin_ass_fail = df_cin_ass[
        (df_cin_ass["CINclosureDate"] < df_cin_ass["AssessmentActualStartDate"])
        | (df_cin_ass["CINclosureDate"] < df_cin_ass["AssessmentAuthorisationDate"])
    ]
    df_cin_ass_fail["ERROR_ID"] = tuple(
        zip(
            df_cin_ass_fail["LAchildID"],
            df_cin_ass_fail["CINdetailsID"],
            df_cin_ass_fail["CINclosureDate"],
        )
    )

    df_cin_issues_ass = (
        df_cin.merge(df_cin_ass_fail, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_ass_issues = (
        df_ass.merge(df_cin_ass_fail, left_on="ROW_ID", right_on="ROW_ID_ass")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # SECTION47 TABLE
    df_cin_47 = df_cin.merge(
        df_47,
        on=["LAchildID", "CINdetailsID"],
        how="left",
        suffixes=["_cin", "_47"],
    )  # both have the DateOfInitialCPC column so suffixes are applied.

    df_cin_47_fail = df_cin_47[
        (df_cin_47["CINclosureDate"] < df_cin_47["S47ActualStartDate"])
        | (df_cin_47["CINclosureDate"] < df_cin_47["DateOfInitialCPC_47"])
    ]
    df_cin_47_fail["ERROR_ID"] = tuple(
        zip(
            df_cin_47_fail["LAchildID"],
            df_cin_47_fail["CINdetailsID"],
            df_cin_47_fail["CINclosureDate"],
        )
    )

    df_cin_issues_47 = (
        df_cin.merge(df_cin_47_fail, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_47_issues = (
        df_47.merge(df_cin_47_fail, left_on="ROW_ID", right_on="ROW_ID_47")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # CHILDPROTECTIONPLANS TABLE

    df_cpp = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINdetailsID": "CINID1",
                "CPPendDate": "27/05/2022",  # fail
            }
        ]
    )

    df_cpp.index.name = "ROW_ID"
    df_cpp.reset_index(inplace=True)

    df_cpp["CPPendDate"] = pd.to_datetime(
        df_cpp["CPPendDate"], format="%d/%m/%Y", errors="coerce"
    )

    df_cin_cpp = df_cin.merge(
        df_cpp,
        on=["LAchildID", "CINdetailsID"],
        how="left",
        suffixes=["_cin", "_cpp"],
    )

    df_cin_cpp_fail = df_cin_cpp[
        df_cin_cpp["CINclosureDate"] < df_cin_cpp["CPPendDate"]
    ]
    df_cin_cpp_fail["ERROR_ID"] = tuple(
        zip(
            df_cin_cpp_fail["LAchildID"],
            df_cin_cpp_fail["CINdetailsID"],
            df_cin_cpp_fail["CINclosureDate"],
        )
    )

    df_cin_issues_cpp = (
        df_cin.merge(df_cin_cpp_fail, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_cpp_issues = (
        df_cpp.merge(df_cin_cpp_fail, left_on="ROW_ID", right_on="ROW_ID_cpp")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_cpp_issues

    # CINPLANDATES TABLE
    df_cin_plan = df_cin.merge(
        df_plan,
        on=["LAchildID", "CINdetailsID"],
        how="left",
        suffixes=["_cin", "_plan"],
    )

    df_cin_plan_fail = df_cin_plan[
        (df_cin_plan["CINclosureDate"] < df_cin_plan["CINPlanStartDate"])
        | (df_cin_plan["CINclosureDate"] < df_cin_plan["CINPlanEndDate"])
    ]
    df_cin_plan_fail["ERROR_ID"] = tuple(
        zip(
            df_cin_47_fail["LAchildID"],
            df_cin_47_fail["CINdetailsID"],
            df_cin_47_fail["CINclosureDate"],
        )
    )

    df_cin_issues_plan = (
        df_cin.merge(df_cin_plan_fail, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_plan_issues = (
        df_plan.merge(df_cin_plan_fail, left_on="ROW_ID", right_on="ROW_ID_plan")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # combine failing cin locations from all tables
    cin_issues_all = pd.concat(
        [
            df_cin_issues_plan,
            df_cin_issues_cpp,
            df_cin_issues_47,
            df_cin_issues_ass,
            df_cin_issues_cin,
        ],
        ignore_index=True,
    )
    unique_cin_issues = cin_issues_all.astype(str).drop_duplicates().index
    df_cin_issues = cin_issues_all.loc[unique_cin_issues].reset_index(drop=True)

    rule_context.push_type_2(
        table=CINdetails,
        columns=[CINclosureDate, DateOfInitialCPC],
        row_df=df_cin_issues,
    )
    rule_context.push_type_2(
        table=Assessments,
        columns=[AssessmentActualStartDate, AssessmentAuthorisationDate],
        row_df=df_ass_issues,
    )
    rule_context.push_type_2(
        table=Section47,
        columns=[S47ActualStartDate, DateOfInitialCPC],
        row_df=df_47_issues,
    )
    rule_context.push_type_2(
        table=ChildProtectionPlans, columns=[CPPendDate], row_df=df_cpp_issues
    )
    rule_context.push_type_2(
        table=CINplanDates,
        columns=[CINPlanStartDate, CINPlanEndDate],
        row_df=df_plan_issues,
    )


def test_validate():
    df_cin = pd.DataFrame(
        [
            # Same CINclosureDate value is maintained throughout table for simplicity when cross-checking with other tables.
            {
                "LAchildID": "child1",
                "CINdetailsID": "CINID1",
                "CINclosureDate": "26/05/2022",
                "DateOfInitialCPC": "26/05/2022",  # pass
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "CINID2",
                "CINclosureDate": "26/05/2022",
                "DateOfInitialCPC": pd.NA,  # ignore. date is absent
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "CINID3",
                "CINclosureDate": "26/05/2022",
                "DateOfInitialCPC": "27/05/2022",  # fail 27/05/2022 is after CINclosureDate
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "CINID4",
                "CINclosureDate": "26/05/2022",
                "DateOfInitialCPC": "26/05/2022",  # pass
            },
        ]
    )

    df_ass = pd.DataFrame(
        [
            # multiple assessments in the same CIN. some pass, some fail.
            {
                "LAchildID": "child1",
                "CINdetailsID": "CINID1",
                "AssessmentActualStartDate": "27/05/2022",  # fail
                "AssessmentAuthorisationDate": "26/05/2022",  # pass
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "CINID1",
                "AssessmentActualStartDate": "26/05/2022",  # pass
                "AssessmentAuthorisationDate": "26/05/2022",  # pass
            },
            # assessments across multiple CIN modules. some pass, some fail.
            {
                "LAchildID": "child1",
                "CINdetailsID": "CINID2",
                "AssessmentActualStartDate": "27/05/2022",  # fail
                "AssessmentAuthorisationDate": "26/05/2022",  # pass
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "CINID3",
                "AssessmentActualStartDate": "26/05/2022",  # pass
                "AssessmentAuthorisationDate": "26/05/2022",  # pass
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "CINID4",
                "AssessmentActualStartDate": pd.NA,  # ignore. date is absent
                "AssessmentAuthorisationDate": pd.NA,  # ignore. date is absent
            },
        ]
    )

    df_47 = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINdetailsID": "CINID1",
                "S47ActualStartDate": "27/05/2022",  # fail. after CINclosureDate.
                "DateOfInitialCPC": "26/05/2022",
            }
        ]
    )

    df_cpp = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINdetailsID": "CINID1",
                "CPPendDate": "27/05/2022",  # fail
            }
        ]
    )

    df_plan = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINdetailsID": "CINID1",
                "CINPlanStartDate": "27/05/2022",  # fail
                "CINPlanEndDate": "26/05/2022",  # pass
            }
        ]
    )

    df_plan["CINPlanStartDate"] = pd.to_datetime(
        df_plan["CINPlanStartDate"], format="%d/%m/%Y", errors="coerce"
    )
    df_plan["CINPlanEndDate"] = pd.to_datetime(
        df_plan["CINPlanEndDate"], format="%d/%m/%Y", errors="coerce"
    )

    df_cpp["CPPendDate"] = pd.to_datetime(
        df_cpp["CPPendDate"], format="%d/%m/%Y", errors="coerce"
    )

    df_47["S47ActualStartDate"] = pd.to_datetime(
        df_47["S47ActualStartDate"], format="%d/%m/%Y", errors="coerce"
    )
    df_47["DateOfInitialCPC"] = pd.to_datetime(
        df_47["DateOfInitialCPC"], format="%d/%m/%Y", errors="coerce"
    )

    df_ass["AssessmentActualStartDate"] = pd.to_datetime(
        df_ass["AssessmentActualStartDate"], format="%d/%m/%Y", errors="coerce"
    )
    df_ass["AssessmentAuthorisationDate"] = pd.to_datetime(
        df_ass["AssessmentAuthorisationDate"], format="%d/%m/%Y", errors="coerce"
    )

    df_cin["CINclosureDate"] = pd.to_datetime(
        df_cin["CINclosureDate"], format="%d/%m/%Y", errors="coerce"
    )
    df_cin["DateOfInitialCPC"] = pd.to_datetime(
        df_cin["DateOfInitialCPC"], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            CINdetails: df_cin,
            Assessments: df_ass,
            Section47: df_47,
            ChildProtectionPlans: df_cpp,
            CINplanDates: df_plan,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 5

    issues = issues_list[0]

    issue_table = issues.table
    assert issue_table == CINdetails

    issue_columns = issues.columns
    assert issue_columns == [
        CINclosureDate,
        DateOfInitialCPC,
    ]

    issue_rows = issues.row_df

    assert len(issue_rows) == 3
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    "CINID1",
                    pd.to_datetime("26/05/2022", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child1",
                    "CINID2",
                    pd.to_datetime("26/05/2022", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child1",
                    "CINID3",
                    pd.to_datetime("26/05/2022", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [2],
            },
        ]
    )

    assert result.definition.code == "8565"
    assert result.definition.message == "Activity shown after a case has been closed"



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
CINdetails = CINTable.CINdetails
Disabilities = CINTable.Disabilities

LAchildID = ChildIdentifiers.LAchildID
PersonBirthDate = ChildIdentifiers.PersonBirthDate
ReferralNFA = CINdetails.ReferralNFA
Disability = Disabilities.Disability


@rule_definition(
    code="8540",
    module=CINTable.ChildCharacteristics,
    message="Child’s disability is missing or invalid (see Disability table)",
    affected_fields=[Disability, PersonBirthDate, ReferralNFA],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_dis = data_container[Disabilities].copy()
    df_ci = data_container[ChildIdentifiers].copy()
    df_cin = data_container[CINdetails].copy()

    df_ci.index.name = "ROW_ID"
    df_dis.index.name = "ROW_ID"
    df_cin.index.name = "ROW_ID"

    df_ci.reset_index(inplace=True)
    df_dis.reset_index(inplace=True)
    df_cin.reset_index(inplace=True)

    # If <PersonBirthDate> (N00066) is present and at least one <ReferralNFA> (N00112) is false or 0,
    # then one or more instances of <Disability> (N00099) must be present,
    # and all instances must be valid code values
    valid_dis = [
        "NONE",
        "MOB",
        "HAND",
        "PC",
        "INC",
        "COMM",
        "LD",
        "HEAR",
        "VIS",
        "BEH",
        "CON",
        "AUT",
        "DDA",
    ]

    df_ci_cin = df_ci.merge(
        df_cin, on="LAchildID", how="left", suffixes=("_ci", "_cin")
    )
    merged_df = df_ci_cin.merge(
        df_dis, on="LAchildID", how="left", suffixes=("", "_dis")
    )

    merged_df = merged_df[~merged_df[Disability].isin(valid_dis)]
    merged_df = merged_df[merged_df[PersonBirthDate].notna()]
    merged_df = merged_df[merged_df[ReferralNFA].isin(["false", "0"])]

    merged_df["ERROR_ID"] = tuple(
        zip(
            merged_df[LAchildID],
            merged_df[PersonBirthDate],
        )
    )

    df_dis_issues = (
        df_dis.merge(merged_df, on="ROW_ID")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_ci_issues = (
        df_ci.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_ci")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID_x"]
        .apply(list)
        .reset_index()
    )
    df_cin_issues = (
        df_cin.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID_x"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=ChildIdentifiers, columns=[PersonBirthDate], row_df=df_ci_issues
    )
    rule_context.push_type_2(
        table=Disabilities, columns=[Disability], row_df=df_dis_issues
    )
    rule_context.push_type_2(
        table=CINdetails, columns=[ReferralNFA], row_df=df_cin_issues
    )


def test_validate():
    sample_ci = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "PersonBirthDate": "01/01/1880",
            },
            {
                "LAchildID": "child2",
                "PersonBirthDate": "01/01/1880",
            },
            {
                "LAchildID": "child3",
                "PersonBirthDate": "01/01/1880",
            },
            {
                "LAchildID": "child4",
                "PersonBirthDate": "01/01/1880",
            },
            {
                "LAchildID": "child5",  # 4 ignore: has no ReferralNFA
                "PersonBirthDate": "01/01/1880",
            },
        ]
    )

    sample_cin = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # considered
                "CINdetailsID": "CINID1",
                "ReferralNFA": "false",
            },
            {
                "LAchildID": "child2",  # considered
                "CINdetailsID": "CINID1",
                "ReferralNFA": "0",
            },
            {
                "LAchildID": "child3",  # considered since one of its ReferralNFA values is false/0
                "CINdetailsID": "CINID1",
                "ReferralNFA": "0",
            },
            {
                "LAchildID": "child3",
                "CINdetailsID": "CINID2",
                "ReferralNFA": "1",
            },
            {
                "LAchildID": "child4",  # 3 ignore ReferralNFA is not false/0
                "CINdetailsID": "CINID1",
                "ReferralNFA": "1",
            },
        ]
    )

    sample_dis = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # 0 fail: Disability should be present
                "Disability": pd.NA,
            },
            {
                "LAchildID": "child2",  # 1 pass
                "Disability": "MOB",
            },
            {
                "LAchildID": "child3",  # 2 fail: Disability should be valid
                "Disability": "notreal",
            },
        ]
    )

    result = run_rule(
        validate,
        {
            ChildIdentifiers: sample_ci,
            Disabilities: sample_dis,
            CINdetails: sample_cin,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 3
    issues = issues_list[1]

    issue_table = issues.table
    assert issue_table == Disabilities

    issue_columns = issues.columns
    assert issue_columns == [Disability]

    issue_rows = issues.row_df
    assert len(issue_rows) == 2
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    "01/01/1880",
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child3",
                    "01/01/1880",
                ),
                "ROW_ID": [2],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "8540"
    assert (
        result.definition.message
        == "Child’s disability is missing or invalid (see Disability table)"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
Sex = ChildIdentifiers.Sex
ExpectedPersonBirthDate = ChildIdentifiers.ExpectedPersonBirthDate


# define characteristics of rule
@rule_definition(
    code="2886Q",
    rule_type=RuleType.QUERY,
    module=CINTable.ChildIdentifiers,
    message="Please check and either amend or provide a reason: Percentage of children with no sex recorded is more than 2% (excluding unborns)",
    affected_fields=[ExpectedPersonBirthDate, ChildIdentifiers],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildIdentifiers]

    # The sum of: the number of child records where <Sex> (N00065) is equal to zero or missing,
    # and the <ExpectedPersonBirthDate> (N00098) is equal to missing, divided by the total number of
    # child records should be less than or equal to 0.02. Validation should be triggered at LA level.

    # get the total number of child records
    num_records = len(df)

    # get the number of child records that fit the specified condition.
    missing_gender = (df[Sex].isna()) | (df[Sex] == "U")
    missing_date = df[ExpectedPersonBirthDate].isna()
    condition = missing_gender & missing_date
    # since the filtered number has to be compared to the original, make a copy of the data.
    df_issues = df.copy()
    df_issues = df_issues[condition]
    num_issues = len(df_issues)

    # calculate
    missing_ratio = num_issues / num_records
    if missing_ratio > 0.02:
        rule_context.push_la_level(
            rule_context.definition.code, rule_context.definition.message
        )
    else:
        pass


def test_validate():
    sample_child_ids = pd.DataFrame(
        [
            {Sex: 0, ExpectedPersonBirthDate: pd.NA},
            {Sex: 1, ExpectedPersonBirthDate: pd.NA},
            {Sex: pd.NA, ExpectedPersonBirthDate: pd.NA},
            {Sex: "U", ExpectedPersonBirthDate: pd.NA},
        ]
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {ChildIdentifiers: sample_child_ids})

    assert result.la_issues == (
        "2886Q",
        "Please check and either amend or provide a reason: Percentage of children with no sex recorded is more than 2% (excluding unborns)",
    )



from functools import wraps
from typing import Callable, Iterable, Optional

from cin_validator.rule_engine.__api import CINTable, RuleDefinition, RuleType


def rule_definition(
    code: str,
    module: CINTable,
    rule_type: RuleType = RuleType.ERROR,
    message: Optional[str] = None,
    affected_fields: Optional[Iterable] = None,
):
    """
    Creates the rule definition for validation rules using RuleDefinition class as a template.

    :param int code: The rule code for each rule.
    :param RuleType-class rule_type: object denoting if the rule is an error or a query.
    :param CINtable-object module: string denoting the module/table affected by a validation rule.
    :param str affected_fields: The fields/columns affected by a validation rule.
    :param str message: The message displayed for each validation rule.
    :returns: RuleDefinition object containing information about validation rules.
    :rtype: RuleDefiniton class object.
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        definition = RuleDefinition(
            code=str(code),
            func=func,
            rule_type=rule_type,
            module=module,
            message=message,
            affected_fields=affected_fields,
        )
        wrapper.__rule_def__ = definition
        return wrapper

    return decorator



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

CINdetails = CINTable.CINdetails
Assessments = CINTable.Assessments

LAchildID = CINdetails.LAchildID
CINdetailsID = CINdetails.CINdetailsID
ReasonForClosure = CINdetails.ReasonForClosure


# define characteristics of rule
@rule_definition(
    # write the rule code here, in place of 2885
    code="8825Q",
    # replace CINdetails with the value in the module column of the excel sheet corresponding to this rule .
    # Note that even if multiple tables are involved, one table will be named in the module column.
    module=CINTable.CINdetails,
    rule_type=RuleType.QUERY,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Please check and either amend data or provide a reason: Reason for Closure code RC8 (case closed after assessment) or RC9 (case closed after assessment, referred to early help) has been returned but there is no assessment present for the episode.",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[
        ReasonForClosure,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    df_ass = data_container[Assessments]
    df_cin = data_container[CINdetails]

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df_ass.index.name = "ROW_ID"
    df_cin.index.name = "ROW_ID"

    # Resetting the index causes the ROW_IDs to become columns of their respective DataFrames
    # so that they can come along when the merge is done.
    df_ass.reset_index(inplace=True)
    df_cin.reset_index(inplace=True)

    # lOGIC
    # If <ReasonforClosure> (N00103) = RC8 or RC9 then at least one <Assessments> module must be present

    df_cin_check = df_cin.copy()

    df_cin_check = df_cin_check[df_cin_check[ReasonForClosure].isin(["RC8", "RC9"])]

    merged_df = df_cin_check.merge(
        df_ass,
        on=[LAchildID, CINdetailsID],
        suffixes=["_cin", "_ass"],
        how="left",
        indicator=True,
    )
    # get modules whose ReasonForClosure is RC8/RC9 but are not found in the assessment table.
    condition = merged_df["_merge"] == "left_only"
    merged_df = merged_df[condition].reset_index()

    # create an identifier for each error instance.
    merged_df["ERROR_ID"] = tuple(zip(merged_df[LAchildID], merged_df[CINdetailsID]))

    # The merges were done on copies of df_ass and df_cin so that the column names in dataframes themselves aren't affected by the suffixes.
    # we can now map the suffixes columns to their corresponding source tables such that the failing ROW_IDs and ERROR_IDs exist per table.
    df_ass_issues = (
        df_ass.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_ass")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_cin_issues = (
        df_cin.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_2(
        table=Assessments, columns=[CINdetailsID], row_df=df_ass_issues
    )
    rule_context.push_type_2(
        table=CINdetails, columns=[ReasonForClosure], row_df=df_cin_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_ass = pd.DataFrame(
        [
            {
                "LAchildID": "child7",
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child3",
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child3",
                "CINdetailsID": "cinID2",
            },
            {
                "LAchildID": "child3",
                "CINdetailsID": "cinID3",
            },
            {
                "LAchildID": "child3",
                "CINdetailsID": "cinID4",
            },
            {
                "LAchildID": "child4",
                "CINdetailsID": "cinID1",
            },
        ]
    )
    sample_cin_details = pd.DataFrame(
        [
            {  # 0 pass
                "LAchildID": "child1",
                "ReasonForClosure": pd.NA,  # 0 ignore
                "CINdetailsID": "cinID1",
            },
            {  # 1 ignore
                "LAchildID": "child1",
                "ReasonForClosure": "EX7",  # 1 ignore
                "CINdetailsID": "cinID2",
            },
            {  # 2
                "LAchildID": "child2",
                "ReasonForClosure": "EX7",  # 2 ignore
                "CINdetailsID": "cinID1",
            },
            {  # 3
                "LAchildID": "child3",
                "ReasonForClosure": "RC8",  # 3 pass. present in assessment table
                "CINdetailsID": "cinID1",
            },
            {  # 4 ignore
                "LAchildID": "child3",
                "ReasonForClosure": "EX7",  # 4 ignore
                "CINdetailsID": "cinID2",
            },
            {  # 5 pass
                "LAchildID": "child3",
                "ReasonForClosure": "RC9",  # 5 pass. present in assessment table
                "CINdetailsID": "cinID3",
            },
            {  # 6 fail
                "LAchildID": "child4",
                "ReasonForClosure": "RC8",  # 6 fail: no assessment recorded.
                "CINdetailsID": "cinID4",
            },
        ]
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            Assessments: sample_ass,
            CINdetails: sample_cin_details,
        },
    )

    # Use .type2_issues to check for the result of .push_type2_issues() which you used above.
    issues_list = result.type2_issues
    assert len(issues_list) == 2
    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)
    # pick any table and check it's values. the tuple in location 1 will contain the CINdetails columns because that's the second thing pushed above.
    issues = issues_list[1]

    # get table name and check it. Replace CINdetails with the name of your table.
    issue_table = issues.table
    assert issue_table == CINdetails

    # check that the right columns were returned. Replace ReasonForClosure with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [ReasonForClosure]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 1 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 1
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child4",
                    "cinID4",
                ),
                "ROW_ID": [6],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace 8825Q with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8825Q"
    assert (
        result.definition.message
        == "Please check and either amend data or provide a reason: Reason for Closure code RC8 (case closed after assessment) or RC9 (case closed after assessment, referred to early help) has been returned but there is no assessment present for the episode."
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

Assessments = CINTable.Assessments
Header = CINTable.Header
AssessmentAuthorisationDate = Assessments.AssessmentAuthorisationDate
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    code="8696",
    module=CINTable.Assessments,
    message="Assessment end date must fall within the census year",
    affected_fields=[AssessmentAuthorisationDate, ReferenceDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[Assessments]
    df_ref = data_container[Header]

    ref_data_series = df_ref[ReferenceDate]
    collection_start, collection_end = make_census_period(ref_data_series)

    # implement rule logic as described by the Github issue. Put the description as a comment above the implementation as shown.

    # <If present <AssessmentAuthorisationDate> (N00160) must be on or between [Start_Of_Census_Year] and <ReferenceDate> (N00603)

    df = df[
        (df[AssessmentAuthorisationDate] < collection_start)
        | (df[AssessmentAuthorisationDate] > collection_end)
    ]

    failing_indices = df.index

    rule_context.push_issue(
        table=Assessments, field=AssessmentAuthorisationDate, row=failing_indices
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.

    miss_auth = pd.to_datetime(
        [
            "01/03/2019",  # 0 fail
            "01/04/2021",  # 1 pass
            "01/10/2022",  # 2 fail
            pd.NA,  # 3 ignored
        ],
        format="%d/%m/%Y",
        errors="coerce",
    )

    fake_auth = pd.DataFrame({AssessmentAuthorisationDate: miss_auth})
    fake_header = pd.DataFrame({ReferenceDate: ["31/03/2022"]})

    # Run rule function passing in our sample data
    result = run_rule(validate, {Assessments: fake_auth, Header: fake_header})

    # The result contains a list of issues encountered
    issues = list(result.issues)
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issues) == 2
    # replace the table and column name as done earlier.
    # The last numbers represent the index values where you expect the sample data to fail the validation check.
    assert issues == [
        IssueLocator(CINTable.Assessments, AssessmentAuthorisationDate, 0),
        IssueLocator(CINTable.Assessments, AssessmentAuthorisationDate, 2),
    ]

    # Check that the rule definition is what you wrote in the context above.

    assert result.definition.code == "8696"
    assert (
        result.definition.message
        == "Assessment end date must fall within the census year"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

CINplanDates = CINTable.CINplanDates
CINPlanStartDate = CINplanDates.CINPlanStartDate
CINPlanEndDate = CINplanDates.CINPlanEndDate
LAchildID = CINplanDates.LAchildID


@rule_definition(
    code="4011",
    module=CINTable.CINplanDates,
    message="CIN Plan End Date earlier than Start Date",
    affected_fields=[CINPlanEndDate, CINPlanStartDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[CINplanDates]
    df.index.name = "ROW_ID"

    # <If present <CINPlanEndDate> (N00690) must be on or after the <CINPlanStartDate> (N00689)
    # Remove all rows with no end date
    df = df[~df[CINPlanEndDate].isna()]

    # Return rows where end date is prior to start dat
    condition1 = df[CINPlanEndDate] < df[CINPlanStartDate]

    # df with all rows meeting the conditions
    df_issues = df[condition1].reset_index()

    link_id = tuple(
        zip(
            df_issues[LAchildID], df_issues[CINPlanEndDate], df_issues[CINPlanStartDate]
        )
    )
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_1(
        table=CINplanDates, columns=[CINPlanEndDate, CINPlanStartDate], row_df=df_issues
    )


def test_validate():
    cin_plan = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINPlanEndDate": "26/05/2000",
                "CINPlanStartDate": "26/05/2000",
            },
            {
                "LAchildID": "child2",
                "CINPlanEndDate": "26/05/2000",
                "CINPlanStartDate": "26/05/2001",
                # fails, start after end
            },
            {
                "LAchildID": "child3",
                "CINPlanEndDate": "26/05/2000",
                "CINPlanStartDate": "26/05/1999",
            },
            {
                "LAchildID": "child4",
                "CINPlanEndDate": "26/05/2000",
                "CINPlanStartDate": pd.NA,
                # pass, no requirement in this rule for a start date
            },
            {
                "LAchildID": "child6",
                "CINPlanEndDate": pd.NA,
                "CINPlanStartDate": pd.NA,
            },
        ]
    )
    cin_plan[CINPlanEndDate] = pd.to_datetime(
        cin_plan[CINPlanEndDate], format="%d/%m/%Y", errors="coerce"
    )
    cin_plan[CINPlanStartDate] = pd.to_datetime(
        cin_plan[CINPlanStartDate], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(validate, {CINplanDates: cin_plan})

    issues = result.type1_issues

    issue_table = issues.table
    assert issue_table == CINplanDates

    issue_columns = issues.columns
    assert issue_columns == [CINPlanEndDate, CINPlanStartDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 1
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child2",
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                    pd.to_datetime("26/05/2001", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "4011"
    assert result.definition.message == "CIN Plan End Date earlier than Start Date"



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py
# Replace ChildIdentifiers with the table name, and LAChildID with the column name you want.

ChildIdentifiers = CINTable.ChildIdentifiers
LAchildID = ChildIdentifiers.LAchildID


# define characteristics of rule
@rule_definition(
    # write the rule code here, in place of '8500'
    code="8500",
    # replace ChildIdentifiers with the value in the module column of the excel sheet corresponding to this rule .
    module=CINTable.ChildIdentifiers,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="LA Child ID missing",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[LAchildID],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # Replace ChildIdentifiers with the name of the table you need.
    df = data_container[ChildIdentifiers]

    # implement rule logic as described by the Github issue. Put the description as a comment above the implementation as shown.

    # <LAchildID> (N00097) must be present
    failing_indices = df[df[LAchildID].isna()].index

    # Replace ChildIdentifiers and LAchildID with the table and column name concerned in your rule, respectively.
    # If there are multiple columns or table, make this sentence multiple times.
    rule_context.push_issue(
        table=ChildIdentifiers, field=LAchildID, row=failing_indices
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    child_identifiers = pd.DataFrame([[1234], [pd.NA], [pd.NA]], columns=[LAchildID])

    # Run rule function passing in our sample data
    result = run_rule(validate, {ChildIdentifiers: child_identifiers})

    # The result contains a list of issues encountered
    issues = list(result.issues)
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issues) == 2
    # replace the table and column name as done earlier.
    # The last numbers represent the index values where you expect the sample data to fail the validation check.
    assert issues == [
        IssueLocator(CINTable.ChildIdentifiers, LAchildID, 1),
        IssueLocator(CINTable.ChildIdentifiers, LAchildID, 2),
    ]

    # Check that the rule definition is what you wrote in the context above.

    # replace '8500' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8500"
    assert result.definition.message == "LA Child ID missing"



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
UPN = ChildIdentifiers.UPN


@rule_definition(
    code="1530",
    module=CINTable.ChildIdentifiers,
    message="UPN invalid (characters 2-4 not a recognised LA code)",
    affected_fields=[UPN],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildIdentifiers]
    """
    If <UPN> (N00001) present then characters 2-4 of <UPN> must be a valid post April 1998 LA code 
    or a recognised ‘pseudo LA’ code 

    001-005, 201-213, 301-320, 330-336, 340-344, 350-359, 370-373, 380-384, 390-394, 420, 660-681, 
    701-708, 800-803, 805-808, 810-813, 815, 816, 820- 823, 825, 826, 830, 831, 835-837, 838-839, 
    840, 841, 845, 846, 850-852, 855-857, 860, 861, 865-896, 908, 909, 916, 919, 921, 925, 
    926, 928, 929, 931, 933, 935-938, 940-941
    """
    LA_list = []
    LA_list.extend("00" + str(x) for x in range(1, 6))
    LA_list.extend(range(201, 214))
    LA_list.extend(range(301, 321))
    LA_list.extend(range(330, 337))
    LA_list.extend(range(340, 345))
    LA_list.extend(range(350, 360))
    LA_list.extend(range(370, 374))
    LA_list.extend(range(380, 385))
    LA_list.extend(range(390, 395))
    LA_list.extend(range(660, 682))
    LA_list.extend(range(701, 709))
    LA_list.extend(range(800, 804))
    LA_list.extend(range(805, 809))
    LA_list.extend(range(810, 814))
    LA_list.extend(range(820, 824))
    LA_list.extend(range(835, 840))
    LA_list.extend(range(850, 853))
    LA_list.extend(range(855, 858))
    LA_list.extend(range(865, 897))
    LA_list.extend(range(935, 939))
    LA_list.extend(range(940, 942))
    LA_list.extend(
        [
            420,
            815,
            816,
            825,
            826,
            830,
            831,
            840,
            841,
            845,
            846,
            860,
            861,
            908,
            909,
            916,
            919,
            921,
            925,
            926,
            928,
            929,
            931,
            933,
        ]
    )

    LA_list = [str(x) for x in LA_list]

    df.reset_index(inplace=True)
    df2 = df[["index", "UPN"]]
    df2 = df2[(df2["UPN"].str.len() == 13) & df2["UPN"].notna()]
    df2["C2_to_C4"] = df2["UPN"].str[1:4]
    df2 = df2[~df2["C2_to_C4"].isin(LA_list)]

    failing_indices = df2.set_index("index").index

    rule_context.push_issue(table=ChildIdentifiers, field=UPN, row=failing_indices)


def test_validate():
    child_identifiers = pd.DataFrame(
        {
            "UPN": [
                # These should pass
                "A38100178301",  # 0 In LA list
                pd.NA,  # 1
                "H003278544154",  # 2 In LA list
                "R34",  # 3 Nonsense
                # These should fail
                "R421962919251",  # 4 Not in LA list
                "X817558133462",  # 5 Not in LA list
                "ASFFAGSVSV123",  # 6 Not in LA list, not numeric.
            ]
        }
    )

    result = run_rule(validate, {ChildIdentifiers: child_identifiers})

    issues = list(result.issues)

    assert len(issues) == 3

    assert issues == [
        IssueLocator(CINTable.ChildIdentifiers, UPN, 4),
        IssueLocator(CINTable.ChildIdentifiers, UPN, 5),
        IssueLocator(CINTable.ChildIdentifiers, UPN, 6),
    ]

    assert result.definition.code == "1530"
    assert (
        result.definition.message
        == "UPN invalid (characters 2-4 not a recognised LA code)"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

Assessments = CINTable.Assessments
LAchildID = Assessments.LAchildID
AssessmentActualStartDate = Assessments.AssessmentActualStartDate
AssessmentAuthorisationDate = Assessments.AssessmentAuthorisationDate


@rule_definition(
    code="8608",
    module=CINTable.Assessments,
    message="Assessment Start Date cannot be later than its End Date",
    affected_fields=[AssessmentActualStartDate, AssessmentAuthorisationDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[Assessments]
    df.index.name = "ROW_ID"

    # If present <AssessmentAuthorisationDate> (N00160) must be on or after the <AssessmentActualStartDate> (N00159)
    condition = df[AssessmentActualStartDate] > df[AssessmentAuthorisationDate]

    df_issues = df[condition].reset_index()

    link_id = tuple(
        zip(
            df_issues[LAchildID],
            df_issues[AssessmentActualStartDate],
            df_issues[AssessmentAuthorisationDate],
        )
    )
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_1(
        table=Assessments,
        columns=[AssessmentActualStartDate, AssessmentAuthorisationDate],
        row_df=df_issues,
    )


def test_validate():
    assessments = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "AssessmentActualStartDate": "26/05/2000",
                "AssessmentAuthorisationDate": "26/05/2000",
            },
            {
                "LAchildID": "child2",
                "AssessmentActualStartDate": "26/05/2000",
                "AssessmentAuthorisationDate": "26/05/2001",
            },
            {
                "LAchildID": "child3",
                "AssessmentActualStartDate": "26/05/2000",
                "AssessmentAuthorisationDate": "26/05/1999",
            },  # 2 error: end is before start
            {
                "LAchildID": "child3",
                "AssessmentActualStartDate": "26/05/2000",
                "AssessmentAuthorisationDate": pd.NA,
            },
            {
                "LAchildID": "child4",
                "AssessmentActualStartDate": "26/05/2000",
                "AssessmentAuthorisationDate": "25/05/2000",
            },  # 4 error: end is before start
            {
                "LAchildID": "child5",
                "AssessmentActualStartDate": pd.NA,
                "AssessmentAuthorisationDate": pd.NA,
            },
        ]
    )

    assessments[AssessmentActualStartDate] = pd.to_datetime(
        assessments[AssessmentActualStartDate], format="%d/%m/%Y", errors="coerce"
    )
    assessments[AssessmentAuthorisationDate] = pd.to_datetime(
        assessments[AssessmentAuthorisationDate], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(validate, {Assessments: assessments})

    issues = result.type1_issues

    issue_table = issues.table
    assert issue_table == Assessments

    issue_columns = issues.columns
    assert issue_columns == [AssessmentActualStartDate, AssessmentAuthorisationDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 2
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child3",
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                    pd.to_datetime("26/05/1999", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [2],
            },
            {
                "ERROR_ID": (
                    "child4",
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                    pd.to_datetime("25/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [4],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "8608"
    assert (
        result.definition.message
        == "Assessment Start Date cannot be later than its End Date"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

ChildIdentifiers = CINTable.ChildIdentifiers
LAchildID = ChildIdentifiers.LAchildID
PersonBirthDate = ChildIdentifiers.PersonBirthDate
ExpectedPersonBirthDate = ChildIdentifiers.ExpectedPersonBirthDate
Sex = ChildIdentifiers.Sex


# define characteristics of rule
@rule_definition(
    # write the rule code here, in place of 8750
    code="8750",
    # replace ChildIdentifiers with the value in the module column of the excel sheet corresponding to this rule .
    module=CINTable.ChildIdentifiers,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Sex must equal U for an unborn child",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[Sex, PersonBirthDate, ExpectedPersonBirthDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    # Replace ChildIdentifiers with the name of the table you need.
    df = data_container[ChildIdentifiers]
    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df.index.name = "ROW_ID"

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # If <ExpectedPersonBirthDate> (N00098) is present and <PersonBirthDate> (N00066) is blank then <Sex> (N00783) must equal “0”
    condition = (
        df[PersonBirthDate].isna()
        & df[ExpectedPersonBirthDate].notna()
        & (df[Sex].astype(str) != "U")
    )
    # get all the data that fits the failing condition. Reset the index so that ROW_ID now becomes a column of df
    df_issues = df[condition].reset_index()

    # SUBMIT ERRORS
    # Generate a unique ID for each instance of an error.
    link_id = tuple(
        zip(
            df_issues[LAchildID],
            df_issues[Sex],
            df_issues[ExpectedPersonBirthDate],
        )
    )

    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    # Ensure that you do not change the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_1(
        table=ChildIdentifiers,
        columns=[Sex, PersonBirthDate, ExpectedPersonBirthDate],
        row_df=df_issues,
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    child_identifiers = pd.DataFrame(
        [
            {  # 0 - Pass - Not unborn
                "LAchildID": "child1",
                "PersonBirthDate": "26/05/2000",
                "ExpectedPersonBirthDate": "26/05/2000",
                "Sex": "M",
            },
            {  # 1 - Pass - Not unborn
                "LAchildID": "child2",
                "PersonBirthDate": "26/05/2000",
                "ExpectedPersonBirthDate": pd.NA,
                "Sex": "F",
            },
            {  # 2 - Pass - Unborn with Sex = U
                "LAchildID": "child3",
                "PersonBirthDate": pd.NA,
                "ExpectedPersonBirthDate": "26/05/1999",
                "Sex": "U",
            },
            {  # 3 - Pass - Not unborn or born! (Not relevant to this rule)
                "LAchildID": "child3",
                "PersonBirthDate": pd.NA,
                "ExpectedPersonBirthDate": pd.NA,
                "Sex": "F",
            },
            {  # 4 - Fail - Unborn with Sex = F
                "LAchildID": "child4",
                "PersonBirthDate": pd.NA,
                "ExpectedPersonBirthDate": "25/05/2000",
                "Sex": "F",
            },
            {  # 5 - Fail - Unborn with Sex = F
                "LAchildID": "child4",
                "PersonBirthDate": pd.NA,
                "ExpectedPersonBirthDate": "25/05/2000",
                "Sex": "M",
            },
        ]
    )
    # if rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    child_identifiers[PersonBirthDate] = pd.to_datetime(
        child_identifiers[PersonBirthDate], format="%d/%m/%Y", errors="coerce"
    )
    child_identifiers[ExpectedPersonBirthDate] = pd.to_datetime(
        child_identifiers[ExpectedPersonBirthDate], format="%d/%m/%Y", errors="coerce"
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {ChildIdentifiers: child_identifiers})

    # Use .type1_issues to check for the result of .push_type1_issues() which you used above.
    issues = result.type1_issues

    # get table name and check it. Replace ChildIdentifiers with the name of your table.
    issue_table = issues.table
    assert issue_table == ChildIdentifiers

    # check that the right columns were returned.
    issue_columns = issues.columns
    assert issue_columns == [Sex, PersonBirthDate, ExpectedPersonBirthDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child4",
                    "F",
                    pd.to_datetime("25/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [4],
            },
            {
                "ERROR_ID": (
                    "child4",
                    "M",
                    pd.to_datetime("25/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [5],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace 8750 with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8750"
    assert result.definition.message == "Sex must equal U for an unborn child"



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

Section47 = CINTable.Section47
S47ActualStartDate = Section47.S47ActualStartDate
DateOfInitialCPC = Section47.DateOfInitialCPC
LAchildID = Section47.LAchildID


# define characteristics of rule
@rule_definition(
    code="8615",
    module=CINTable.Section47,
    message="Section 47 Enquiry Start Date must be present and cannot be later than the date of the initial Child Protection Conference",
    affected_fields=[S47ActualStartDate, DateOfInitialCPC],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[Section47]

    # Within a <Section47> group, if <DateOfInitialCPC> (N00110) is present then <S47ActualStartDate> (N00148) must be present and on or before the <DateOfInitialCPC> (N00110)
    df.index.name = "ROW_ID"

    df.query(
        "(S47ActualStartDate > DateOfInitialCPC) or (DateOfInitialCPC.notna() and S47ActualStartDate.isna())",
        inplace=True,
    )

    df_issues = df.reset_index()

    link_id = tuple(
        zip(
            df_issues[LAchildID],
            df_issues[S47ActualStartDate],
            df_issues[DateOfInitialCPC],
        )
    )
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_1(
        table=Section47,
        columns=[S47ActualStartDate, DateOfInitialCPC],
        row_df=df_issues,
    )


def test_validate():
    IDS_are = [
        "AAAAAAAA",
        "BBBBBBBBB",
        "CCCCCCCCCCC",
        "DDDDDDDDD",
        "EEEE",
        "FFFFFFFFF",
        "GGGGGGGGGG",
        "HHHH",
    ]
    S47start = [
        "01-01-2020",
        "01-02-2020",
        "01-03-2020",
        "15-01-2020",
        pd.NA,
        "01-07-2020",
        "15-01-2020",
        pd.NA,
    ]
    ICPCstart = [
        "01-01-2020",
        "01-01-2020",  #  Fails as ICPC before S47.
        "01-03-2020",
        "17-01-2020",
        pd.NA,
        "01-01-2020",  #  Fails as ICPC before S47.
        "15-01-2020",
        "01-01-2020",  #  Fails as ICPC with no S47.
    ]
    fake_dataframe = pd.DataFrame(
        {
            "LAchildID": IDS_are,
            "S47ActualStartDate": S47start,
            "DateOfInitialCPC": ICPCstart,
        }
    )

    fake_dataframe["S47ActualStartDate"] = pd.to_datetime(
        fake_dataframe["S47ActualStartDate"], format=r"%d-%m-%Y", errors="coerce"
    )
    fake_dataframe["DateOfInitialCPC"] = pd.to_datetime(
        fake_dataframe["DateOfInitialCPC"], format=r"%d-%m-%Y", errors="coerce"
    )

    result = run_rule(validate, {Section47: fake_dataframe})

    issues = result.type1_issues

    issue_table = issues.table
    assert issue_table == Section47

    issue_columns = issues.columns
    assert issue_columns == [S47ActualStartDate, DateOfInitialCPC]

    issue_rows = issues.row_df

    assert len(issue_rows) == 3

    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "BBBBBBBBB",
                    pd.to_datetime("01-02-2020", format=r"%d-%m-%Y", errors="coerce"),
                    pd.to_datetime("01-01-2020", format=r"%d-%m-%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "FFFFFFFFF",
                    pd.to_datetime("01-07-2020", format=r"%d-%m-%Y", errors="coerce"),
                    pd.to_datetime("01-01-2020", format=r"%d-%m-%Y", errors="coerce"),
                ),
                "ROW_ID": [5],
            },
            {
                "ERROR_ID": (
                    "HHHH",
                    pd.NaT,
                    pd.to_datetime("01-01-2020", format=r"%d-%m-%Y", errors="coerce"),
                ),
                "ROW_ID": [7],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "8615"
    assert (
        result.definition.message
        == "Section 47 Enquiry Start Date must be present and cannot be later than the date of the initial Child Protection Conference"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

CINplanDates = CINTable.CINplanDates
LAchildID = CINplanDates.LAchildID
CINPlanStartDate = CINplanDates.CINPlanStartDate
CINPlanEndDate = CINplanDates.CINPlanEndDate

Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


@rule_definition(
    code="4014",
    module=CINTable.CINplanDates,
    message="CIN Plan data contains overlapping dates",
    affected_fields=[
        CINPlanStartDate,
        CINPlanEndDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_cinp = data_container[CINplanDates].copy()
    df_cinp2 = data_container[CINplanDates].copy()

    df_cinp.index.name = "ROW_ID"
    df_cinp2.index.name = "ROW_ID"

    df_cinp.reset_index(inplace=True)
    df_cinp2.reset_index(inplace=True)

    df_ref = data_container[Header]
    ref_date_series = df_ref[ReferenceDate]

    collection_start, reference_date = make_census_period(ref_date_series)

    # Where more than one <CINplanDates> group is included,
    # the <CINPlanStartDate> (N00105) of each group cannot fall within either:
    #   a) <CINPlanStartDate> (N00105) to <CINPlanEndDate> (N00115), or
    #   b) <CINPlanStartDate> (N00105) and <ReferenceDate> if <CINPlanEndDate> (N00115) is not present
    # of any other group
    #
    # Issues dfs should return rows where CINPlanStartDate is between another CINPlanStartDate and CINPlanEndDate (or ReferenceDate)

    #  Create dataframes which only have rows with CIN plans, and which should have one plan per row.
    df_cinp = df_cinp[df_cinp[CINPlanStartDate].notna()]
    df_cinp2 = df_cinp2[df_cinp2[CINPlanStartDate].notna()]

    #  Merge tables to test for overlaps
    df_merged = df_cinp.merge(
        df_cinp2,
        on=["LAchildID"],
        how="left",
        suffixes=("_cinp", "_cinp2"),
    )

    # Use CINPlanStartDate to identify a CIN plan. Exclude rows where the ROW_ID is the same on both sides to prevent a plan from being compared with itself.
    df_merged = df_merged[df_merged["ROW_ID_cinp"] != df_merged["ROW_ID_cinp2"]]

    # Determine whether CINplanStart overlaps with another CINplan period of the same child.
    cinp_started_after_start = (
        df_merged["CINPlanStartDate_cinp"] >= df_merged["CINPlanStartDate_cinp2"]
    )
    cinp_started_before_end = (
        df_merged["CINPlanStartDate_cinp"] <= df_merged["CINPlanEndDate_cinp2"]
    ) & df_merged["CINPlanEndDate_cinp2"].notna()
    cinp_started_before_refdate = (
        df_merged["CINPlanStartDate_cinp"] <= reference_date
    ) & df_merged["CINPlanEndDate_cinp2"].isna()

    df_merged = df_merged[
        cinp_started_after_start
        & (cinp_started_before_end | cinp_started_before_refdate)
    ].reset_index()

    # create an identifier for each error instance.
    # In this case, the rule is checked for each CINPlanStartDate, in each CPplanDates group (differentiated by CP dates), in each child (differentiated by LAchildID)
    df_merged["ERROR_ID"] = tuple(
        zip(
            df_merged[LAchildID],
            df_merged["CINPlanStartDate_cinp"],
            df_merged["CINPlanStartDate_cinp2"],
        )
    )

    # The merges were done on copies of cinp_df so that the column names in dataframes themselves aren't affected by the suffixes.
    # we can now map the suffixes columns to their corresponding source tables such that the failing ROW_IDs and ERROR_IDs exist per table.
    df_cinp_issues = (
        df_cinp.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cinp")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    df_cinp2_issues = (
        df_cinp2.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cinp2")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_3(
        table=CINplanDates, columns=[CINPlanStartDate], row_df=df_cinp_issues
    )
    rule_context.push_type_3(
        table=CINplanDates,
        columns=[CINPlanStartDate, CINPlanEndDate],
        row_df=df_cinp2_issues,
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2001"}]  # the census start date here will be 01/04/2000
    )

    sample_cinp = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # 0 Pass
                "CINPlanStartDate": "26/05/2000",
                "CINPlanEndDate": "26/10/2000",
            },
            {
                "LAchildID": "child1",  # 1 Fail because of 0
                "CINPlanStartDate": "26/08/2000",
                "CINPlanEndDate": "26/12/2000",
            },
            {
                "LAchildID": "child2",  # 2 Pass
                "CINPlanStartDate": "26/05/2000",
                "CINPlanEndDate": "25/10/2000",
            },
            {
                "LAchildID": "child2",  # 3 Pass
                "CINPlanStartDate": "26/10/2000",
                "CINPlanEndDate": "26/12/2000",
            },
            {
                "LAchildID": "child3",  # 4 Pass
                "CINPlanStartDate": "26/05/2000",
                "CINPlanEndDate": pd.NA,
            },
            {
                "LAchildID": "child3",  # 5 Fail
                "CINPlanStartDate": "26/08/2000",
                "CINPlanEndDate": "26/10/2000",
            },
            {
                "LAchildID": "child4",  # 6 Pass
                "CINPlanStartDate": "26/10/2000",
                "CINPlanEndDate": "31/03/2001",
            },
            {
                "LAchildID": "child4",  # 7 Fail
                "CINPlanStartDate": "31/03/2001",
                "CINPlanEndDate": pd.NA,
            },
            {
                "LAchildID": "child5",  # 8 Fail
                "CINPlanStartDate": "31/03/2001",
                "CINPlanEndDate": "31/04/2001",
            },
            {
                "LAchildID": "child5",  # 9 Fail
                "CINPlanStartDate": "31/03/2001",
                "CINPlanEndDate": "31/04/2001",
            },
        ]
    )

    # If rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_cinp[CINPlanStartDate] = pd.to_datetime(
        sample_cinp[CINPlanStartDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_cinp["CINPlanEndDate"] = pd.to_datetime(
        sample_cinp["CINPlanEndDate"], format="%d/%m/%Y", errors="coerce"
    )
    sample_header[ReferenceDate] = pd.to_datetime(
        sample_header[ReferenceDate], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            CINplanDates: sample_cinp,
            Header: sample_header,
        },
    )

    issues_list = result.type3_issues
    assert len(issues_list) == 2

    issues = issues_list[0]

    issue_table = issues.table
    assert issue_table == CINplanDates

    issue_columns = issues.columns
    assert issue_columns == [CINPlanStartDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 4

    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    pd.to_datetime("26/08/2000", format="%d/%m/%Y", errors="coerce"),
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child3",
                    pd.to_datetime("26/08/2000", format="%d/%m/%Y", errors="coerce"),
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [5],
            },
            {
                "ERROR_ID": (
                    "child4",
                    pd.to_datetime("31/03/2001", format="%d/%m/%Y", errors="coerce"),
                    pd.to_datetime("26/10/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [7],
            },
            {
                "ERROR_ID": (
                    "child5",
                    pd.to_datetime("31/03/2001", format="%d/%m/%Y", errors="coerce"),
                    pd.to_datetime("31/03/2001", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [8, 9],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "4014"
    assert result.definition.message == "CIN Plan data contains overlapping dates"



"""
Rule number: 8870Q
Module: Section 47
Rule details: Where present, the <InitialCPCtarget> (N00109) should not be a Saturday, Sunday

Rule message: Please check: The Target Date for Initial Child Protection Conference should not be a weekend

"""
from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    RuleType,
    rule_definition,
)
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

Section47 = CINTable.Section47
InitialCPCtarget = Section47.InitialCPCtarget


# define characteristics of rule
@rule_definition(
    code="8870Q",
    module=CINTable.Section47,
    rule_type=RuleType.QUERY,
    message="Please check and either amend or provide a reason: The Target Date for Initial Child Protection Conference should not be a weekend",
    affected_fields=[InitialCPCtarget],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[Section47]

    # <InitialCPCtarget> (N00109) should not be a Saturday, Sunday
    # Convert column to date format
    df[InitialCPCtarget] = pd.to_datetime(
        df[InitialCPCtarget], format="%d-%m-%Y", errors="coerce"
    )

    # .weekday() returns the integer value for each day (0-6) with weekends being 5 and 6
    failing_indices = df[df[InitialCPCtarget].dt.weekday >= 5].index

    rule_context.push_issue(
        table=Section47, field=InitialCPCtarget, row=failing_indices
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    child_identifiers = pd.DataFrame(
        [["17-06-2021"], ["09-10-2022"], ["14-03-2020"]], columns=[InitialCPCtarget]
    )  # 9/10 is a Sunday and 14/3 is a Saturday

    # Run rule function passing in our sample data
    result = run_rule(validate, {Section47: child_identifiers})

    # The result contains a list of issues encountered
    issues = list(result.issues)
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issues) == 2
    # replace the table and column name as done earlier.
    # The last numbers represent the index values where you expect the sample data to fail the validation check.
    assert issues == [
        IssueLocator(CINTable.Section47, InitialCPCtarget, 1),
        IssueLocator(CINTable.Section47, InitialCPCtarget, 2),
    ]

    # Check that the rule definition is what you wrote in the context above.

    assert result.definition.code == "8870Q"
    assert (
        result.definition.message
        == "Please check and either amend or provide a reason: The Target Date for Initial Child Protection Conference should not be a weekend"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py
# Replace ChildIdentifiers with the table name, and GenderCurrent with the column name you want.

ChildIdentifiers = CINTable.ChildIdentifiers
GenderCurrent = ChildIdentifiers.GenderCurrent


# define characteristics of rule
@rule_definition(
    # write the rule code here, in place of 8500
    code="4180",
    # replace ChildIdentifiers with the value in the module column of the excel sheet corresponding to this rule .
    module=CINTable.ChildIdentifiers,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Gender is missing",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[GenderCurrent],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # Replace ChildIdentifiers with the name of the table you need.
    df = data_container[ChildIdentifiers]

    # implement rule logic as described by the Github issue. Put the description as a comment above the implementation as shown.

    valid_gender_codes = ["1", "2", "0", "9"]

    # <GenderCurrent> (N00097) must be present and valid

    failing_indices = df[
        df[GenderCurrent].isna()
        | (~df[GenderCurrent].astype("str").isin(valid_gender_codes))
    ].index

    # Replace ChildIdentifiers and GenderCurrent with the table and column name concerned in your rule, respectively.
    # If there are multiple columns or table, make this sentence multiple times.
    rule_context.push_issue(
        table=ChildIdentifiers, field=GenderCurrent, row=failing_indices
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    child_identifiers = pd.DataFrame(
        [1, pd.NA, 7, "Male", 2, 0, 9, "9"], columns=[GenderCurrent]
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {ChildIdentifiers: child_identifiers})

    # The result contains a list of issues encountered
    issues = list(result.issues)
    # replace 3 with the number of failing points you expect from the sample data.
    assert len(issues) == 3
    # replace the table and column name as done earlier.
    # The last numbers represent the index values where you expect the sample data to fail the validation check.
    assert issues == [
        IssueLocator(CINTable.ChildIdentifiers, GenderCurrent, 1),
        IssueLocator(CINTable.ChildIdentifiers, GenderCurrent, 2),
        IssueLocator(CINTable.ChildIdentifiers, GenderCurrent, 3),
    ]

    # Check that the rule definition is what you wrote in the context above.

    # replace '4180' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "4180"
    assert result.definition.message == "Gender is missing"



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

Assessments = CINTable.Assessments
AssessmentActualStartDate = Assessments.AssessmentActualStartDate
LAchildID = Assessments.LAchildID
Assdetailsid = Assessments.CINdetailsID

CINdetails = CINTable.CINdetails
CINreferralDate = CINdetails.CINreferralDate
LAchildID = CINdetails.LAchildID
CINdetailsID = CINdetails.CINdetailsID


@rule_definition(
    code="1103",
    module=CINTable.Assessments,
    message="The assessment start date cannot be before the referral date",
    affected_fields=[
        AssessmentActualStartDate,
        CINreferralDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_ass = data_container[Assessments].copy()
    df_refs = data_container[CINdetails].copy()

    df_ass.index.name = "ROW_ID"
    df_refs.index.name = "ROW_ID"

    df_ass.reset_index(inplace=True)
    df_refs.reset_index(inplace=True)

    # Where present, the <AssessmentActualStartDate> (N00159) should be on or after the <CINReferralDate> (N00100)
    # Issues dfs should return rows where Assessment Start Date is less than the Referral Start Date
    df_ass = df_ass[df_ass[AssessmentActualStartDate].notna()]
    df_refs = df_refs[df_refs[CINreferralDate].notna()]

    #  Merge tables to get corresponding Assessment group and referrals
    df_merged = df_ass.merge(
        df_refs,
        left_on=["LAchildID", "CINdetailsID"],
        right_on=["LAchildID", "CINdetailsID"],
        how="left",
        suffixes=("_ass", "_refs"),
    )

    #  Get rows where Assessment Start Date is less than the Referral Start Date
    condition = df_merged[AssessmentActualStartDate] < df_merged[CINreferralDate]
    df_merged = df_merged[condition].reset_index()

    df_merged["ERROR_ID"] = tuple(
        zip(
            df_merged[LAchildID],
            df_merged[AssessmentActualStartDate],
            df_merged[CINreferralDate],
        )
    )

    # The merges were done on copies of fs_ass and df_refs so that the column names in dataframes themselves aren't affected by the suffixes.
    # we can now map the suffixes columns to their corresponding source tables such that the failing ROW_IDs and ERROR_IDs exist per table.
    df_ass_issues = (
        df_ass.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_ass")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_refs_issues = (
        df_refs.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_refs")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_2(
        table=Assessments, columns=[AssessmentActualStartDate], row_df=df_ass_issues
    )
    rule_context.push_type_2(
        table=CINdetails, columns=[CINreferralDate], row_df=df_refs_issues
    )


def test_validate():
    sample_ass = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "AssessmentActualStartDate": "30/06/2021",  # Fails as referral date is after assessment start
                "CINdetailsID": "CIN1",
            },
            {
                "LAchildID": "child2",
                "AssessmentActualStartDate": "10/09/2021",  #  Passes as assessment starts after referal start date
                "CINdetailsID": "CIN2",
            },
            {
                "LAchildID": "child3",
                "AssessmentActualStartDate": pd.NA,  # Ignored as no Assessment Date recorded
                "CINdetailsID": "CIN3",
            },
            {
                "LAchildID": "child4",
                "AssessmentActualStartDate": "01/12/2021",  # Fails as assessment starts after referral start date
                "CINdetailsID": "CIN4",
            },
            {
                "LAchildID": "child5",
                "AssessmentActualStartDate": "10/02/2022",  # Fails as no Referral Start Date recorded
                "CINdetailsID": "CIN5",
            },
        ]
    )
    sample_refs = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # Fails
                "CINreferralDate": "01/07/2021",
                "CINdetailsID": "CIN1",
            },
            {
                "LAchildID": "child2",  # Passes
                "CINreferralDate": "01/09/2021",
                "CINdetailsID": "CIN2",
            },
            {
                "LAchildID": "child3",  # Ignored
                "CINreferralDate": "26/05/2000",
                "CINdetailsID": "CIN3",
            },
            {
                "LAchildID": "child4",  # Fails
                "CINreferralDate": "10/12/2021",
                "CINdetailsID": "CIN4",
            },
            {
                "LAchildID": "child5",  # Ignored
                "CINreferralDate": pd.NA,
                "CINdetailsID": "CIN5",
            },
        ]
    )

    sample_ass[AssessmentActualStartDate] = pd.to_datetime(
        sample_ass[AssessmentActualStartDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_refs["CINreferralDate"] = pd.to_datetime(
        sample_refs["CINreferralDate"], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            Assessments: sample_ass,
            CINdetails: sample_refs,
        },
    )

    # Type 2 rule.
    issues_list = result.type2_issues
    assert len(issues_list) == 2
    issues = issues_list[1]

    issue_table = issues.table
    assert issue_table == CINdetails

    issue_columns = issues.columns
    assert issue_columns == [CINreferralDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 2
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",  # ChildID
                    # Assessment Date
                    pd.to_datetime("30/06/2021", format="%d/%m/%Y", errors="coerce"),
                    # Referral date
                    pd.to_datetime("01/07/2021", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child4",  # ChildID
                    # Assessmwent date
                    pd.to_datetime("01/12/2021", format="%d/%m/%Y", errors="coerce"),
                    # Referral date
                    pd.to_datetime("10/12/2021", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [3],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "1103"
    assert (
        result.definition.message
        == "The assessment start date cannot be before the referral date"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

ChildIdentifiers = CINTable.ChildIdentifiers
Header = CINTable.Header
PersonBirthDate = ChildIdentifiers.PersonBirthDate
ReferenceDate = Header.ReferenceDate


@rule_definition(
    code="8520",
    module=CINTable.ChildIdentifiers,
    message="Date of Birth is after data collection period (must be on or before the end of the census period)",
    affected_fields=[PersonBirthDate, ReferenceDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildIdentifiers]
    df_ref = data_container[Header]

    ref_data_series = df_ref[ReferenceDate]
    collection_start, collection_end = make_census_period(ref_data_series)

    # <PersonBirthDate> (N00066) must be on or before <ReferenceDate> (N00603) or null

    df = df[(df[PersonBirthDate] > collection_end)]

    failing_indices = df.index

    rule_context.push_issue(
        table=ChildIdentifiers, field=PersonBirthDate, row=failing_indices
    )


def test_validate():
    test_dobs = pd.to_datetime(
        [
            "01/06/2021",  # 0 pass
            "01/06/2022",  # 1 fail
            "31/08/2021",  # 2 pass
            "31/12/2022",  # 3 fail
            pd.NA,  # 4 ignored
        ],
        format="%d/%m/%Y",
        errors="coerce",
    )

    fake_dobs = pd.DataFrame({PersonBirthDate: test_dobs})
    fake_header = pd.DataFrame({ReferenceDate: ["31/03/2022"]})

    result = run_rule(validate, {ChildIdentifiers: fake_dobs, Header: fake_header})

    issues = list(result.issues)
    assert len(issues) == 2
    assert issues == [
        IssueLocator(CINTable.ChildIdentifiers, PersonBirthDate, 1),
        IssueLocator(CINTable.ChildIdentifiers, PersonBirthDate, 3),
    ]

    assert result.definition.code == "8520"
    assert (
        result.definition.message
        == "Date of Birth is after data collection period (must be on or before the end of the census period)"
    )



import datetime
import importlib
import json
import os
import xml.etree.ElementTree as ET
from pathlib import Path

import click
import pytest

from cin_validator import cin_validator


@click.group()
def cli():
    pass


@cli.command(name="list")
@click.option(
    "--ruleset",
    "-r",
    default="cin2025_26",
    help="Which ruleset to use, e.g. cin2025_26",
)
def list_cmd(ruleset):
    """
    List all rules in a ruleset.

    Call using:
    python -m cin_validator list

    :param str ruleset: The name of a CIN validation ruleset, used to select rules for validating data.
    :returns: A list of validation rules in the given ruleset.
    :rtype: list
    """
    module = importlib.import_module(f"cin_validator.rules.{ruleset}")
    ruleset_registry = getattr(module, "registry")
    for _, rule in ruleset_registry.items():
        click.echo(f"{rule.code}\t{rule.message} ({rule.rule_type.name})")


@cli.command(name="run")
@click.argument("filename", type=click.File("rt"), required=True)
@click.option(
    "--ruleset",
    "-r",
    default="cin2024_25",
    help="Which ruleset to use, e.g. cin2025_26",
)
@click.option("--select", "-s", default=None)
@click.option("--output/--no_output", "-o/-no", default=False)
def run_all(filename: str, ruleset, select, output):
    """
    Used to run all of a set of validation rules on input data.

    CLI command:
    python -m cin_validator run <filepath_to_data>

    Can be used to validate data, via the command line interface, for a given rule set.
    Runs with the cin2025_26 ruleset as standard.

    :param str filename: Refers to the filepath of data to be validated.
    :param str ruleset: The folder name of the validation rules to run input data against.
    :param select: specify the rules that should be run. CLI works with a single string only.
    :param bool output: If true, produces csv output of error report, if False (default)
        does not.
    :returns: DataFrame report of errors using selected validation rules, also output as
        JSON when output is True.
    :rtype: DataFrame, JSON
    """

    fulltree = ET.parse(filename)
    root = fulltree.getroot()

    raw_data = cin_validator.convert_data(root)
    data_files = cin_validator.process_data(raw_data)

    # get rules based on specified year.
    module = importlib.import_module(f"cin_validator.rules.{ruleset}")
    ruleset_registry = getattr(module, "registry")

    validator = cin_validator.CinValidator(
        data_files, ruleset_registry, selected_rules=select
    )

    full_issue_df = validator.full_issue_df

    if output:
        validator.user_report.to_csv("user_report.csv")

    click.echo(full_issue_df)
    # # click.echo(validator.multichild_issues)
    # click.echo(validator.data_files["Assessments"])


@cli.command(name="test")
@click.argument("rule", required=False)
@click.option(
    "--ruleset",
    "-r",
    default="cin2025_26",
    help="Which ruleset to use, e.g. cin2025_26",
)
def test_cmd(rule, ruleset):
    """
    Test all (or individual) rules in a ruleset. Note: tests the code
    for the rule, this is not used for validating data.

    Allows use of the CLI to test a ruleset or individual rules against the
    pytest written in each of their files. Useful for bugfixing. Defaults
    to the cin2022_23 ruleset.

    Can be called to test all rules using:
    python -m cin_validator test
    To test individual rules:
    python -m cin_validator <rulecode>
    For example:
    python -m cin_validator test 8875

    :param str rule: Used to specify an individual rule to test.
    :param str ruleset: Use to give the name of a set of validation rules to test
        (defaults to cin2025_26).
    :returns: Pytest output in terminal of rules passing and failing.
    :rtype: Pytest output in terminal.
    """

    module = importlib.import_module(f"cin_validator.rules.{ruleset}")
    module_folder = Path(module.__file__).parent

    ruleset_registry = getattr(module, "registry")

    if rule:
        rule = str(rule)
        # when rule code is specified, test specific rule.
        rule_def = ruleset_registry.get(rule)
        if not rule_def:
            # if the get returns a <NoneType>
            click.secho(f"Rule {rule} not found.", err=True, fg="red")
            return 1
        test_files = [os.path.join(module_folder, f"rule_{rule}.py")]
    else:
        # else test all rules.
        test_files = [
            str(p.absolute())
            for p in module_folder.glob("*.py")
            if p.stem != "__init__"
        ]

    pytest.main(test_files)


@cli.command(name="xmltocsv")
@click.argument("filename", type=click.Path(), required=True)
def cli_converter(filename: str):
    """
    Converts XML to CSV at selected filepath. Does not require XML to be validated against validation rules and does not validate against rules.
    Called using:
    python -m cin_validator xmltocsv <filepath>

    :param str filename: filename (or path) of XML file to convert to CSV.
    :returns: CSV of XML input into output_csvs directory (which will be created
        if it doesn't already exist).
    :rtype: CSVs (multiple).

    """
    if Path(filename).exists():
        fulltree = ET.parse(filename)
        root = fulltree.getroot()

        cin_tables_dict = cin_validator.convert_data(root)
        for k, v in cin_tables_dict.items():
            filepath = Path(f"output_csvs/{k}.csv")
            filepath.parent.mkdir(parents=True, exist_ok=True)
            v.to_csv(filepath)
    else:
        click.echo(f"{filename} can't be found, have you entered it correctly?")


@cli.command(name="timer")
@click.argument("filepath", type=str, required=True)
def timer(filepath):
    st = datetime.datetime.now()
    os.system(f"python -m cin_validator run {filepath}")
    et = datetime.datetime.now()
    time_elapsed = et - st
    click.echo(f"Time to run: {time_elapsed}")


if __name__ == "__main__":
    cli()



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

ChildProtectionPlans = CINTable.ChildProtectionPlans
LAchildID = ChildProtectionPlans.LAchildID
CPPID = ChildProtectionPlans.CPPID
CPPstartDate = ChildProtectionPlans.CPPstartDate
CPPendDate = ChildProtectionPlans.CPPendDate

Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    # write the rule code here
    code="8940",
    # replace ChildProtectionPlans with the value in the module column of the excel sheet corresponding to this rule .
    # Note that even if multiple tables are involved, one table will be named in the module column.
    module=CINTable.ChildProtectionPlans,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Child Protection Plan data contains overlapping dates",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[
        CPPstartDate,
        CPPendDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    # Replace ChildProtectionPlans with the name of the table you need.
    df_cpp = data_container[ChildProtectionPlans].copy()
    df_cpp2 = data_container[ChildProtectionPlans].copy()

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df_cpp.index.name = "ROW_ID"
    df_cpp2.index.name = "ROW_ID"

    # Resetting the index causes the ROW_IDs to become columns of their respective DataFrames
    # so that they can come along when the merge is done.
    df_cpp.reset_index(inplace=True)
    df_cpp2.reset_index(inplace=True)

    # ReferenceDate exists in the header table so we get header table too.
    df_ref = data_container[Header]
    ref_date_series = df_ref[ReferenceDate]

    # the make_census_period function generates the start and end date so that you don't have to do it each time.
    collection_start, reference_date = make_census_period(ref_date_series)

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # Where more than one <ChildProtectionPlans> group is included,
    # the <CPPstartDate> (N00105) of each group cannot fall within either:
    #   a) <CPPstartDate> (N00105) to <CPPendDate> (N00115), or
    #   b) <CPPstartDate> (N00105) and <ReferenceDate> if <CPPendDate> (N00115) is not present
    # of any other group
    #
    # Issues dfs should return rows where CPPstartDate is between another CPPstartDate and CPPendDate (or ReferenceDate)

    #  Create dataframes which only have rows with CP plans, and which should have one plan per row.
    df_cpp = df_cpp[df_cpp[CPPstartDate].notna()]
    df_cpp2 = df_cpp2[df_cpp2[CPPstartDate].notna()]

    #  Merge tables to test for overlaps
    df_merged = df_cpp.merge(
        df_cpp2,
        on=["LAchildID"],
        how="left",
        suffixes=("_cpp", "_cpp2"),
    )

    # Exclude rows where the CPPID is the same on both sides
    df_merged = df_merged[(df_merged["CPPID_cpp"] != df_merged["CPPID_cpp2"])]

    # Determine whether CPP overlaps another CPP
    cpp_started_after_start = (
        df_merged["CPPstartDate_cpp"] >= df_merged["CPPstartDate_cpp2"]
    )
    cpp_started_before_end = (
        df_merged["CPPstartDate_cpp"] <= df_merged["CPPendDate_cpp2"]
    ) & df_merged["CPPendDate_cpp2"].notna()
    cpp_started_before_refdate = (
        df_merged["CPPstartDate_cpp"] <= reference_date
    ) & df_merged["CPPendDate_cpp2"].isna()

    df_merged = df_merged[
        cpp_started_after_start & (cpp_started_before_end | cpp_started_before_refdate)
    ].reset_index()

    # create an identifier for each error instance.
    # In this case, the rule is checked for each CPPstartDate, in each CPplanDates group (differentiated by CP dates), in each child (differentiated by LAchildID)
    df_merged["ERROR_ID"] = tuple(
        zip(df_merged[LAchildID], df_merged["CPPID_cpp"], df_merged["CPPID_cpp2"])
    )

    # The merges were done on copies of cpp_df so that the column names in dataframes themselves aren't affected by the suffixes.
    # we can now map the suffixes columns to their corresponding source tables such that the failing ROW_IDs and ERROR_IDs exist per table.
    df_cpp_issues = (
        df_cpp.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cpp")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_cpp2_issues = (
        df_cpp2.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cpp2")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_3(
        table=ChildProtectionPlans, columns=[CPPstartDate], row_df=df_cpp_issues
    )
    rule_context.push_type_3(
        table=ChildProtectionPlans,
        columns=[CPPstartDate, CPPendDate],
        row_df=df_cpp2_issues,
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2001"}]  # the census start date here will be 01/04/2000
    )

    sample_cpp = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # 0 Pass
                "CPPstartDate": "26/05/2000",
                "CPPendDate": "26/10/2000",
                "CPPID": "cinID1",
            },
            {
                "LAchildID": "child1",  # 1 Fail
                "CPPstartDate": "26/08/2000",
                "CPPendDate": "26/12/2000",
                "CPPID": "cinID12",
            },
            {
                "LAchildID": "child2",  # 2 Pass
                "CPPstartDate": "26/05/2000",
                "CPPendDate": "25/10/2000",
                "CPPID": "cinID2",
            },
            {
                "LAchildID": "child2",  # 3 Pass
                "CPPstartDate": "26/10/2000",
                "CPPendDate": "26/12/2000",
                "CPPID": "cinID22",
            },
            {
                "LAchildID": "child3",  # 4 Pass
                "CPPstartDate": "26/05/2000",
                "CPPendDate": pd.NA,
                "CPPID": "cinID3",
            },
            {
                "LAchildID": "child3",  # 5 Fail
                "CPPstartDate": "26/08/2000",
                "CPPendDate": "26/10/2000",
                "CPPID": "cinID32",
            },
            {
                "LAchildID": "child4",  # 6 Pass
                "CPPstartDate": "26/10/2000",
                "CPPendDate": "31/03/2001",
                "CPPID": "cinID4",
            },
            {
                "LAchildID": "child4",  # 7 Fail
                "CPPstartDate": "31/03/2001",
                "CPPendDate": pd.NA,
                "CPPID": "cinID42",
            },
        ]
    )

    # If rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_cpp[CPPstartDate] = pd.to_datetime(
        sample_cpp[CPPstartDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_cpp["CPPendDate"] = pd.to_datetime(
        sample_cpp["CPPendDate"], format="%d/%m/%Y", errors="coerce"
    )
    sample_header[ReferenceDate] = pd.to_datetime(
        sample_header[ReferenceDate], format="%d/%m/%Y", errors="coerce"
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            ChildProtectionPlans: sample_cpp,
            Header: sample_header,
        },
    )

    # Use .type2_issues to check for the result of .push_type2_issues() which you used above.
    issues_list = result.type3_issues
    assert len(issues_list) == 2
    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)
    # pick any table and check it's values. the tuple in location 1 will contain the Reviews columns because that's the second thing pushed above.
    issues = issues_list[0]

    # get table name and check it. Replace Reviews with the name of your table.
    issue_table = issues.table
    assert issue_table == ChildProtectionPlans

    # check that the right columns were returned. Replace CPPreviewDate  with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [CPPstartDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 3 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 3

    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    "cinID12",
                    "cinID1",
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child3",
                    "cinID32",
                    "cinID3",
                ),
                "ROW_ID": [5],
            },
            {
                "ERROR_ID": (
                    "child4",
                    "cinID42",
                    "cinID4",
                ),
                "ROW_ID": [7],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace 2885 with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8940"
    assert (
        result.definition.message
        == "Child Protection Plan data contains overlapping dates"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

CINplanDates = CINTable.CINplanDates
LAchildID = CINplanDates.LAchildID
CINPlanStartDate = CINplanDates.CINPlanStartDate
Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


@rule_definition(
    code="4010",
    module=CINTable.CINplanDates,
    message="CIN Plan start date is missing or out of data collection period",
    affected_fields=[CINPlanStartDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[CINplanDates]
    df_ref = data_container[Header]

    ref_data_series = df_ref[ReferenceDate]
    collection_start, reference_date = make_census_period(ref_data_series)

    # Where a <CINPlanDates> module is present, <CINPlanStartDate> (N00689) must be present and on or before the <ReferenceDate> (N00603)
    # condition states that there must be a value in CINPlanStartDate and that value must be after the reference_date.
    # As such any NULL values for CINPlanStartDate wil be flagged with the error message
    df = df[(df[CINPlanStartDate].isna() | (df[CINPlanStartDate] > reference_date))]
    failing_indices = df.index

    rule_context.push_issue(
        table=CINplanDates, field=CINPlanStartDate, row=failing_indices
    )


def test_validate():
    cin_start = pd.to_datetime(
        # Create some sample data such that some values pass the validation and some fail.
        [
            "2022-04-25",  # Fail - CINPlanStartDate is after the reference_date of 31st March 2022
            "2022-03-01",  # Pass - CINPlanStartDate is before the reference_date of 31st March 2022
            "2022-12-25",  # Fail - CINPlanStartDate is after the reference_date of 31st March 2022
            "2021-04-27",  # Pass - CINPlanStartDate is before the reference_date of 31st March 2022
            "2021-11-21",  # Pass - CINPlanStartDate is before the reference_date of 31st March 2022
            "2021-08-20",  # Pass - CINPlanStartDate is before the reference_date of 31st March 2022
            "2020-04-17",  # Pass - CINPlanStartDate is before the reference_date of 31st March 2022
            "1999-01-30",  # Pass - CINPlanStartDate is before the reference_date of 31st March 2022
            pd.NA,  # Fail - CINPlanStartDate is blank and is therefore missing.
        ],
        format="%Y/%m/%d",
        errors="coerce",
    )
    CINids = [
        "CINID1",
        "CINID2",
        "CINID3",
        "CINID4",
        "CINID5",
        "CINID6",
        "CINID7",
        "CINID8",
        "CINID9",
    ]

    fake_cin_start = pd.DataFrame({CINPlanStartDate: cin_start, "CINdetailsID": CINids})
    fake_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2022"}]
    )  # the reference date will be 31/03/2022

    fake_header["ReferenceDate"] = pd.to_datetime(
        fake_header["ReferenceDate"], format="%d/%m/%Y"
    )

    result = run_rule(validate, {CINplanDates: fake_cin_start, Header: fake_header})

    issues = list(result.issues)

    assert len(issues) == 3
    assert issues == [
        IssueLocator(CINTable.CINplanDates, CINPlanStartDate, 0),
        IssueLocator(CINTable.CINplanDates, CINPlanStartDate, 2),
        IssueLocator(CINTable.CINplanDates, CINPlanStartDate, 8),
    ]

    assert result.definition.code == "4010"
    assert (
        result.definition.message
        == "CIN Plan start date is missing or out of data collection period"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

CINdetails = CINTable.CINdetails
CINclosureDate = CINdetails.CINclosureDate
Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    code="8620",
    module=CINTable.CINdetails,
    message="CIN Closure Date present and does not fall within the Census year",
    affected_fields=[CINclosureDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[CINdetails]
    df_ref = data_container[Header]

    # ReferenceDate exists in the header table so we get header table too.
    ref_date_series = df_ref[ReferenceDate]

    # the make_census_period function generates the start and end date so that you don't have to do it each time.
    collection_start, collection_end = make_census_period(ref_date_series)

    # implement rule logic as described by the Github issue. Put the description as a comment above the implementation as shown.

    # If <CINclosureDate> (N00102) is present, it must be within [Period_of_Census]
    df = df[df[CINclosureDate].notna()]
    df = df[
        ~(
            (df[CINclosureDate] >= collection_start)
            & (df[CINclosureDate] <= collection_end)
        )
    ]
    failing_indices = df.index

    rule_context.push_issue(table=CINdetails, field=CINclosureDate, row=failing_indices)


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    fake_header = pd.DataFrame(
        {ReferenceDate: ["31/03/2022"]}  # the census start date here will be 01/04/2021
    )
    fake_cinclosure = pd.DataFrame(
        [
            {
                CINclosureDate: "01/03/2019"
            },  # 0 fail: March 1st is before April 1st, 2021. It is out of range
            {
                CINclosureDate: "10/04/2021"
            },  # 1 pass: April 10th is within April 1st, 2021 to March 31st, 2022.
            {
                CINclosureDate: "01/10/2022"
            },  # 2 fail: October 1st is after March 31st, 2022. It is out of range
            {CINclosureDate: "01/04/2021"},  # Pass, first day of census period
            {CINclosureDate: "31/03/2021"},  # Pass, last day of census period
        ]
    )

    # if date columns are involved, the validate function will be expecting them as dates so convert before passing them in.
    fake_cinclosure[CINclosureDate] = pd.to_datetime(
        fake_cinclosure[CINclosureDate], format="%d/%m/%Y", errors="coerce"
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {CINdetails: fake_cinclosure, Header: fake_header})

    # The result contains a list of issues encountered
    issues = list(result.issues)
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issues) == 2
    # replace the table and column name as done earlier.
    # The last numbers represent the index values where you expect the sample data to fail the validation check.
    assert issues == [
        IssueLocator(CINTable.CINdetails, CINclosureDate, 0),
        IssueLocator(CINTable.CINdetails, CINclosureDate, 2),
    ]

    # Check that the rule definition is what you wrote in the context above.

    # replace '8620' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8620"
    assert (
        result.definition.message
        == "CIN Closure Date present and does not fall within the Census year"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
PersonDeathDate = ChildIdentifiers.PersonDeathDate
LAchildID = ChildIdentifiers.LAchildID

CINplanDates = CINTable.CINplanDates
LAchildID = CINplanDates.LAchildID
CINPlanStartDate = CINplanDates.CINPlanStartDate


@rule_definition(
    code="4008",
    module=CINTable.ChildIdentifiers,
    message="CIN Plan shown as starting after the child’s Date of Death.",
    affected_fields=[
        PersonDeathDate,
        CINPlanStartDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_ci = data_container[ChildIdentifiers].copy()
    df_cpd = data_container[CINplanDates].copy()

    df_ci.index.name = "ROW_ID"
    df_cpd.index.name = "ROW_ID"

    df_ci.reset_index(inplace=True)
    df_cpd.reset_index(inplace=True)

    # If <PersonDeathDate> (N00108) is present, then <CINPlanStartDate> (N00689) must be on or before <PersonDeathDate> (N00108)
    df_ci = df_ci[df_ci[PersonDeathDate].notna()]
    df_cpd = df_cpd[df_cpd[CINPlanStartDate].notna()]

    df_merged = df_ci.merge(
        df_cpd,
        left_on=["LAchildID"],
        right_on=["LAchildID"],
        how="left",
        suffixes=("_ci", "_cpd"),
    )

    condition = df_merged[PersonDeathDate] < df_merged[CINPlanStartDate]
    df_merged = df_merged[condition].reset_index()

    df_merged["ERROR_ID"] = tuple(
        zip(
            df_merged[LAchildID],
            df_merged[PersonDeathDate],
        )
    )

    df_cpp_issues = (
        df_ci.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_ci")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_reviews_issues = (
        df_cpd.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cpd")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=ChildIdentifiers, columns=[PersonDeathDate], row_df=df_cpp_issues
    )
    rule_context.push_type_2(
        table=CINplanDates, columns=[CINPlanStartDate], row_df=df_reviews_issues
    )


def test_validate():
    sample_ci = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "PersonDeathDate": "26/05/2000",  # Passes same date
            },
            {
                "LAchildID": "child2",
                "PersonDeathDate": "27/06/2002",  # Passes
            },
            {
                "LAchildID": "child3",
                "PersonDeathDate": "07/02/2001",  # Passes
            },
            {
                "LAchildID": "child4",
                "PersonDeathDate": "26/05/2000",  # Passes
            },
            {
                "LAchildID": "child5",
                "PersonDeathDate": "26/05/2000",  # Fails, death before CIN plan starts
            },
            {
                "LAchildID": "child6",
                "PersonDeathDate": pd.NA,  # Passes
            },
            {
                "LAchildID": "child7",
                "PersonDeathDate": "14/03/2001",  # Passes
            },
        ]
    )
    sample_cpd = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINPlanStartDate": "26/05/2000",
            },
            {
                "LAchildID": "child2",
                "CINPlanStartDate": "26/05/2000",
            },
            {
                "LAchildID": "child3",
                "CINPlanStartDate": "26/05/2000",
            },
            {
                "LAchildID": "child4",
                "CINPlanStartDate": "25/05/2000",
            },
            {
                "LAchildID": "child5",
                "CINPlanStartDate": "27/05/2000",
            },
            {
                "LAchildID": "child6",
                "CINPlanStartDate": "26/05/2000",
            },
            {
                "LAchildID": "child7",
                "CINPlanStartDate": pd.NA,
            },
        ]
    )

    sample_ci["PersonDeathDate"] = pd.to_datetime(
        sample_ci["PersonDeathDate"], format="%d/%m/%Y", errors="coerce"
    )
    sample_cpd["CINPlanStartDate"] = pd.to_datetime(
        sample_cpd["CINPlanStartDate"], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            ChildIdentifiers: sample_ci,
            CINplanDates: sample_cpd,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 2
    issues = issues_list[1]

    issue_table = issues.table
    assert issue_table == CINplanDates

    issue_columns = issues.columns
    assert issue_columns == [CINPlanStartDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 1
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child5",
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [4],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "4008"
    assert (
        result.definition.message
        == "CIN Plan shown as starting after the child’s Date of Death."
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    RuleType,
    rule_definition,
)
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
FormerUPN = ChildIdentifiers.FormerUPN


@rule_definition(
    code="1560Q",
    module=CINTable.ChildIdentifiers,
    rule_type=RuleType.QUERY,
    message="Please check and either amend or provide a reason: Former UPN wrongly formatted",
    affected_fields=[FormerUPN],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildIdentifiers]

    # <FormerUPN> (N00002) where present should be in the correct format, as specified in the data table

    # Note, there are multiple types of former UPN, there are Temporary UPNs which end in a letter, and
    # those where a child is assigned a UPN but then another is identified for them having been used previously.
    # If this was only a check for temporary UPNs, it would check that the last character was a letter. However, it checks more generally.

    #  filter rows of df where the UPN column doesn't have Na/NaN values
    df = df.loc[df[FormerUPN].notna()]

    # Flag locations where FormerUPN is not 13 characters long
    check_length = df[FormerUPN].str.len() != 13
    # Flag locations where FormerUPN's last twelve characters do not form a full digit
    digit_within = ~df[FormerUPN].str[1:].str.isdigit()
    # Flag locations where FormerUPN's first character is not a letter
    check_edges = ~df[FormerUPN].str[0].str.isalpha()

    failing_indices = df[check_length | digit_within | check_edges].index

    rule_context.push_issue(
        table=ChildIdentifiers, field=FormerUPN, row=failing_indices
    )


def test_validate():
    child_identifiers = pd.DataFrame(
        [
            {FormerUPN: pd.NA},  # 0 ignore
            {FormerUPN: "X987654321231"},  # 1 pass
            {FormerUPN: "X0000y0000007"},  # 2 fail non-alphabet within
            {FormerUPN: "X98721238"},  # 3 wrong length
            {FormerUPN: "E000215119000"},
            {FormerUPN: "X987654321231"},  # 1 pass
            {FormerUPN: "E000215119000"},
        ]
    )

    result = run_rule(validate, {ChildIdentifiers: child_identifiers})

    issues = list(result.issues)
    assert len(issues) == 2
    assert issues == [
        IssueLocator(CINTable.ChildIdentifiers, FormerUPN, 2),
        IssueLocator(CINTable.ChildIdentifiers, FormerUPN, 3),
    ]

    assert result.definition.code == "1560Q"
    assert (
        result.definition.message
        == "Please check and either amend or provide a reason: Former UPN wrongly formatted"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
CINplanDates = CINTable.CINplanDates

LAchildID = ChildIdentifiers.LAchildID
PersonDeathDate = ChildIdentifiers.PersonDeathDate
CINPlanEndDate = CINplanDates.CINPlanEndDate


@rule_definition(
    code="4009Q",
    module=CINTable.ChildIdentifiers,
    rule_type=RuleType.QUERY,
    message="CIN Plan cannot end after the child’s Date of Death",
    affected_fields=[
        PersonDeathDate,
        CINPlanEndDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_ci = data_container[ChildIdentifiers].copy()
    df_cin = data_container[CINplanDates].copy()

    df_ci.index.name = "ROW_ID"
    df_cin.index.name = "ROW_ID"

    df_ci.reset_index(inplace=True)
    df_cin.reset_index(inplace=True)

    # If <PersonDeathDate> (N00108) is present, then <CINPlanEndDate> (N00690) must be on or before <PersonDeathDate> (N00108)
    df_ci = df_ci[df_ci["PersonDeathDate"].notna()]

    merged_df = df_ci.merge(
        df_cin,
        on=[
            LAchildID,
        ],
        suffixes=["_ci", "_cin"],
    )

    condition = merged_df[PersonDeathDate] < merged_df[CINPlanEndDate]

    merged_df = merged_df[condition].reset_index()

    merged_df["ERROR_ID"] = tuple(
        zip(
            merged_df[LAchildID],
            merged_df[PersonDeathDate],
        )
    )

    df_ci_issues = (
        df_ci.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_ci")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    df_cin_issues = (
        df_cin.merge(merged_df, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=ChildIdentifiers, columns=[PersonDeathDate], row_df=df_ci_issues
    )
    rule_context.push_type_2(
        table=CINplanDates, columns=[CINPlanEndDate], row_df=df_cin_issues
    )


def test_validate():
    sample_ci = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # Fail
                "PersonDeathDate": "01/01/2000",
            },
            {
                "LAchildID": "child2",  # Pass
                "PersonDeathDate": "01/01/2000",
            },
            {
                "LAchildID": "child3",  # Pass
                "PersonDeathDate": pd.NA,
            },
            {
                "LAchildID": "child4",  # Fail
                "PersonDeathDate": "01/01/2000",
            },
        ]
    )
    sample_cin = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINPlanEndDate": "01/01/2001",
            },
            {
                "LAchildID": "child2",
                "CINPlanEndDate": pd.NA,
            },
            {
                "LAchildID": "child3",
                "CINPlanEndDate": "01/01/2000",
            },
            {
                "LAchildID": "child4",
                "CINPlanEndDate": "01/01/2001",
            },
        ]
    )
    # if rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_ci[PersonDeathDate] = pd.to_datetime(
        sample_ci[PersonDeathDate], format="%d/%m/%Y", errors="coerce"
    )

    sample_cin["CINPlanEndDate"] = pd.to_datetime(
        sample_cin["CINPlanEndDate"], format="%d/%m/%Y", errors="coerce"
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            ChildIdentifiers: sample_ci,
            CINplanDates: sample_cin,
        },
    )

    # Use .type2_issues to check for the result of .push_type2_issues() which you used above.
    issues_list = result.type2_issues
    assert len(issues_list) == 2
    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)
    # pick any table and check it's values. the tuple in location 1 will contain the Section47 columns because that's the second thing pushed above.
    issues = issues_list[0]

    # get table name and check it. Replace Section47 with the name of your table.
    issue_table = issues.table
    assert issue_table == ChildIdentifiers

    # check that the right columns were returned. Replace DateOfInitialCPC  with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [PersonDeathDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    pd.to_datetime("01/01/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child4",
                    pd.to_datetime("01/01/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [3],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace 4009Q with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "4009Q"
    assert (
        result.definition.message
        == "CIN Plan cannot end after the child’s Date of Death"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    RuleType,
    rule_definition,
)
from cin_validator.test_engine import run_rule

CINplanDates = CINTable.CINplanDates
CINPlanStartDate = CINplanDates.CINPlanStartDate
CINPlanEndDate = CINplanDates.CINPlanEndDate
LAchildID = CINplanDates.LAchildID


@rule_definition(
    code="4012Q",
    rule_type=RuleType.QUERY,
    module=CINTable.CINplanDates,
    message="Please check and either amend or provide a reason: CIN Plan shown as starting and ending on the same day",
    affected_fields=[CINPlanStartDate, CINPlanEndDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[CINplanDates]

    # Within a <CINPlanDates> group, <CINPlanStartDate> (N00689) should not be the same as the <CINPlanEndDate> (N00690)
    df.index.name = "ROW_ID"
    df = df[df["CINPlanStartDate"] == df["CINPlanEndDate"]]

    df_issues = df.reset_index()

    link_id = tuple(
        zip(
            df_issues[LAchildID], df_issues[CINPlanStartDate], df_issues[CINPlanEndDate]
        )
    )
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    rule_context.push_type_1(
        table=CINplanDates, columns=[CINPlanStartDate, CINPlanEndDate], row_df=df_issues
    )


def test_validate():
    IDS_are = [
        "AAAAAAAA",
        "BBBBBBBBB",
        "CCCCCCCCCCC",
        "DDDDDDDDD",
        "EEEE",
        "FFFFFFFFF",
        "GGGGGGGGGG",
        "HHHH",
    ]
    starts = [
        "01-01-2020",
        "01-02-2020",
        "01-03-2020",
        "15-01-2020",
        pd.NA,
        "01-07-2020",
        "15-01-2020",
        pd.NA,
    ]
    ends = [
        "01-01-2020",
        "01-01-2020",
        "01-03-2020",
        "17-01-2020",
        pd.NA,
        "01-01-2020",
        "15-01-2020",
        "01-01-2020",
    ]
    #  Fails rows 0, 2, 6.
    fake_dataframe = pd.DataFrame(
        {"LAchildID": IDS_are, "CINPlanStartDate": starts, "CINPlanEndDate": ends}
    )

    fake_dataframe[CINPlanStartDate] = pd.to_datetime(
        fake_dataframe[CINPlanStartDate], format=r"%d-%m-%Y", errors="coerce"
    )
    fake_dataframe[CINPlanEndDate] = pd.to_datetime(
        fake_dataframe[CINPlanEndDate], format=r"%d-%m-%Y", errors="coerce"
    )

    result = run_rule(validate, {CINplanDates: fake_dataframe})

    issues = result.type1_issues

    issue_table = issues.table
    assert issue_table == CINplanDates

    issue_columns = issues.columns
    assert issue_columns == [CINPlanStartDate, CINPlanEndDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 3
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "AAAAAAAA",
                    pd.to_datetime("01-01-2020", format=r"%d-%m-%Y", errors="coerce"),
                    pd.to_datetime("01-01-2020", format=r"%d-%m-%Y", errors="coerce"),
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "CCCCCCCCCCC",
                    pd.to_datetime("01-03-2020", format=r"%d-%m-%Y", errors="coerce"),
                    pd.to_datetime("01-03-2020", format=r"%d-%m-%Y", errors="coerce"),
                ),
                "ROW_ID": [2],
            },
            {
                "ERROR_ID": (
                    "GGGGGGGGGG",
                    pd.to_datetime("15-01-2020", format=r"%d-%m-%Y", errors="coerce"),
                    pd.to_datetime("15-01-2020", format=r"%d-%m-%Y", errors="coerce"),
                ),
                "ROW_ID": [6],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "4012Q"
    assert (
        result.definition.message
        == "Please check and either amend or provide a reason: CIN Plan shown as starting and ending on the same day"
    )



from typing import Callable

from cin_validator.rule_engine import RuleContext, RuleDefinition


def run_rule(rule_func: RuleDefinition, datasets: dict) -> RuleContext:
    ctx = RuleContext(rule_func.__rule_def__)
    rule_func(datasets, ctx)
    return ctx



from pathlib import Path

from cin_validator.rules.ruleset_utils import extract_validator_functions

files = Path(__file__).parent.glob("*.py")
registry = extract_validator_functions(files)

__all__ = ["registry"]



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

CINdetails = CINTable.CINdetails
LAchildID = CINdetails.LAchildID
CINdetailsID = CINdetails.CINdetailsID
CINclosureDate = CINdetails.CINclosureDate
ReferralNFA = CINdetails.ReferralNFA

Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    # write the rule code here
    code="8815",
    # replace CINdetails with the value in the module column of the excel sheet corresponding to this rule .
    # Note that even if multiple tables are involved, one table will be named in the module column.
    module=CINTable.CINdetails,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="More than one open CIN Details episode (a module with no CIN Closure Date) has been provided for this child and case is not a referral with no further action.",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[
        ReferralNFA,
        CINclosureDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    df = data_container[CINdetails]
    # Rename and reset index
    df.index.name = "ROW_ID"
    df.reset_index(inplace=True)

    # If more than one <CINdetails> module is present for the same child, then there must only be one where:
    # <CINclosureDate> (N00102) is missing; and
    # <ReferralNFA> (N00112) = false or 0

    # Notes: a) <CINclosureDate> (N00102) can be present on all modules,
    # b) there may be more than one module with no <CINclosureDate> where <ReferralNFA> (N00112) = true/1.
    falsezero = ["false", "0"]
    df_check = df.copy()
    df_check = df_check[
        df_check[CINclosureDate].isna() & (df_check[ReferralNFA].isin(falsezero))
    ]

    # Convert NAs to 1 and count by child
    df_check[CINclosureDate].fillna(1, inplace=True)

    df_check = df_check.groupby([LAchildID])[CINclosureDate].count().reset_index()

    # Find where there is more than 1 open end date
    # Note, this is done differently to other similar rules, because they compare within one module, and this compares all open modules of the same type.
    fail_list = df_check[df_check[CINclosureDate] > 1]["LAchildID"].tolist()

    # Some of these lines are technically redundant and could be compressed but I've kept them to keep the structure of the rules
    df_bad = df[(df[LAchildID].isin(fail_list)) & df[CINclosureDate].isna()]

    issue_ids = tuple(zip(df_bad[LAchildID], df_bad[CINdetailsID]))

    df["ERROR_ID"] = tuple(zip(df[LAchildID], df[CINdetailsID]))
    df_issues = df[df["ERROR_ID"].isin(issue_ids)]

    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_3(
        table=CINdetails, columns=[CINclosureDate], row_df=df_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2001"}]  # the census start date here will be 01/04/2000
    )

    sample_cin = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # 0 Pass
                "CINreferralDate": "26/05/2000",
                "CINclosureDate": pd.NA,
                "CINdetailsID": "cinID1",
                "ReferralNFA": "true",
            },
            {
                "LAchildID": "child1",  # Pass
                "CINreferralDate": "26/08/2000",
                "CINclosureDate": pd.NA,
                "CINdetailsID": "cinID12",
                "ReferralNFA": "0",
            },
            {
                "LAchildID": "child2",  # 2 Pass
                "CINreferralDate": "26/05/2000",
                "CINclosureDate": "25/10/2000",
                "CINdetailsID": "cinID2",
                "ReferralNFA": "true",
            },
            {
                "LAchildID": "child2",  # 3 Pass
                "CINreferralDate": "26/10/2000",
                "CINclosureDate": "26/12/2000",
                "CINdetailsID": "cinID22",
                "ReferralNFA": "true",
            },
            {
                "LAchildID": "child3",  # 4 Fail
                "CINreferralDate": "26/05/2000",
                "CINclosureDate": pd.NA,
                "CINdetailsID": "cinID3",
                "ReferralNFA": "0",
            },
            {
                "LAchildID": "child3",  # 5 Fail
                "CINreferralDate": "26/08/2000",
                "CINclosureDate": pd.NA,
                "CINdetailsID": "cinID32",
                "ReferralNFA": "0",
            },
            {
                "LAchildID": "child4",  # 6 Fail
                "CINreferralDate": "26/10/2000",
                "CINclosureDate": pd.NA,
                "CINdetailsID": "cinID4",
                "ReferralNFA": "false",
            },
            {
                "LAchildID": "child4",  # 7 Fail
                "CINreferralDate": "31/03/2001",
                "CINclosureDate": pd.NA,
                "CINdetailsID": "cinID42",
                "ReferralNFA": "0",
            },
        ]
    )

    # If rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_cin["CINreferralDate"] = pd.to_datetime(
        sample_cin["CINreferralDate"], format="%d/%m/%Y", errors="coerce"
    )
    sample_cin[CINclosureDate] = pd.to_datetime(
        sample_cin[CINclosureDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_header[ReferenceDate] = pd.to_datetime(
        sample_header[ReferenceDate], format="%d/%m/%Y", errors="coerce"
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            CINdetails: sample_cin,
            Header: sample_header,
        },
    )

    # Use .type3_issues to check for the result of .push_type3_issues() which you used above.
    issues_list = result.type3_issues
    assert len(issues_list) == 1
    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)
    # pick any table and check it's values. the tuple in location 0 will contain the CINdetails columns because that's the second thing pushed above.
    issues = issues_list[0]

    # get table name and check it. Replace CINdetails with the name of your table.
    issue_table = issues.table
    assert issue_table == CINdetails

    # check that the right columns were returned. Replace CINclosureDate with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [CINclosureDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    assert len(issue_rows) == 4

    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child3",
                    "cinID3",
                ),
                "ROW_ID": [4],
            },
            {
                "ERROR_ID": (
                    "child3",
                    "cinID32",
                ),
                "ROW_ID": [5],
            },
            {
                "ERROR_ID": (
                    "child4",
                    "cinID4",
                ),
                "ROW_ID": [6],
            },
            {
                "ERROR_ID": (
                    "child4",
                    "cinID42",
                ),
                "ROW_ID": [7],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace '8815' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8815"
    assert (
        result.definition.message
        == "More than one open CIN Details episode (a module with no CIN Closure Date) has been provided for this child and case is not a referral with no further action."
    )
